"""
Connectors
----------

*Connectors* in WHAD are dedicated classes used to connect an *application*
to a WHAD compatible hardware (any compatible *device*) in order to provide
a set of features. We can see a *connector* as a role applied to a device,
usually related to a *domain* (or *wireless protocol*), that exposes methods
to perform various tasks that rely on a subset of commands supported by the
hardware.

*Connectors* shall ensure the device they are linked to does support the
target domain and a mimimal set of commands, and can tailor its behavior
depending on the capabilities of the hardware. If a *connector* is linked
to a device that either does not support the *domain* this *connector* is
supposed to operate or lacks specific *commands*, a
:py::class:`whad.exceptions.UnsupportedDomain` exception or a
:py:class:`whad.exceptions.UnsupportedCapability` may be raised.

Default connector features
~~~~~~~~~~~~~~~~~~~~~~~~~~

WHAD provides a default connector class, :py:class:`whad.device.connector.Connector`,
that implements a set of features out-of-the-box:

- Packet and message sniffing and processing
- Event notification mechanism
- Synchronous mode

Sniffing packet and messages could be useful to implement packet sniffers or
intercept some specific events like disconnection of the linked hardware device.
Most of the time this feature is used to sniff packets related to a target domain.
The :py:function:`whad.device.connector.Connector.sniff` method is specifically
tailored for this use. When not sniffing, packets received from the hardware device
are forwarded to the connector's packet processing methods than can be overriden by
inheriting classes.

By default, the default connector class provides methods to add and remove custom
event listeners (:py:function:`whad.device.connector.Connector.add_listener` and
:py:function:`whad.device.connector.Connector.remove_listener`), and an additional
method to send an event to the registered listeners (:py:function:`whad.device.connector.Connector.notify`).

Last but not least, the provided *synchronous mode* will disable packet forwarding
and save all received packets in a reception queue, waiting for the application to
retrieve and process them. Service messages will still be processed by the *connector*,
in order to handle any device disconnection or other unexpected event that may occur.
When this *synchronous mode* is disabled, every unprocessed packet stored in the
reception queue are automatically forwarded to the connector's packet processing
methods, and will be then dispatched to the corresponding handlers.
"""

import logging
import contextlib
from time import time
from queue import Queue, Empty
from threading import Thread, Lock, Event as ThreadEvent
from typing import Generator, Callable, Union, List, Optional

from scapy.packet import Packet

from whad.helpers import message_filter
from whad.hub import ProtocolHub
from whad.hub.message import AbstractPacket, AbstractEvent, HubMessage
from whad.hub.generic.cmdresult import CommandResult, Success
from whad.exceptions import WhadDeviceError, WhadDeviceDisconnected, \
    RequiredImplementation, UnsupportedDomain

from .device import Device, DeviceEvt, Disconnected, MessageReceived

logger = logging.getLogger(__name__)

class Event:
    """Generic connector event class.
    """

    def __str__(self):
        return "Event()"

    def __repr__(self):
        return "Event()"

class Notification:
    """Generic connector notification class."""

    def __str__(self) -> str:
        """String representation for notification."""
        return "Notification()"

    def __repr__(self) -> str:
        """Python representation."""
        return str(self)


class ConnIoThread(Thread):
    """Connector's background thread processing events from interface.

    This thread reads the associated connector's event queue and dispatch
    events to the `on_iface_event()` connector's method. Using a separate thread
    to process events sent by the interface avoids concurrency issues.
    """
    def __init__(self, connector: 'Connector'):
        """Connector thread intialization.

        :param connector: Connector associated with this I/O thread.
        :type connector: whad.device.connector.Connector
        """
        super().__init__()
        self.__connector = connector
        self.__canceled = False
        self.daemon = True

    def cancel(self):
        """Cancel IO thread"""
        self.__canceled = True

    def run(self):
        """Connector IO thread main task.
        """
        while not self.__canceled:
            try:
                # Retrieve pending event
                with self.__connector.get_event() as evt:
                    # Let connector process this message
                    self.__connector.on_device_event(evt)
            except Empty:
                pass

class Connector:
    """
    Interface connector.

    A connector creates a link between a device and a protocol controller.
    """

    # Synchronous modes
    SYNC_MODE_OFF = 0
    SYNC_MODE_PKT = 1
    SYNC_MODE_ALL = 2

    def __init__(self, device: Optional[Device] = None):
        """
        Constructor.

        Link the device with this connector, and this connector with the
        provided device.

        :param device: Device to be used with this connector.
        :type device: Device
        """
        self.__device = None
        self.set_device(device)
        if self.__device is not None:
            self.__device.set_connector(self)

        # Interface stall
        self.__stalled = False

        # Packet callbacks
        self.__callbacks_lock = Lock()
        self.__reception_callbacks = {}
        self.__transmission_callbacks = {}
        self.__error_callbacks = []

        # Connector lock mode
        self.__locked = False
        self.__locked_pdus = Queue()
        self.__lock = Lock()

        # Synchronous mode (not enabled by default)
        self.__sync_mode = Connector.SYNC_MODE_OFF

        # Synchronous events (device event + messages)
        self.__sync_events = Queue()

        # Queue holding events coming from our interface
        self.__events = Queue()

        # Create a background thread for message processing
        self.__io_thread = ConnIoThread(self)

        # Event listeners
        self.__listeners = []

        # Interface disconnection
        self.__disconnected = ThreadEvent()

        # Start background thread (start processing messages)
        self.__io_thread.start()


    def attach_error_callback(self, callback, context=None):
        '''Attach an error callback to this connector.

        :param callback: function handling errors.
        :param context:  context object to pass to the error handling function.
        :returns: Boolean indicating if the callback has been successfully attached.
        '''
        # add callbacks to error callbacks
        with self.__callbacks_lock:
            self.__error_callbacks.append(
                (callback, context)
            )

    def on_error(self, error):
        '''Triggers a call to the device connector error handling registered callback(s).
        '''
        if len(self.__error_callbacks) > 0:
            with self.__callbacks_lock:
                # Duplicate error callbacks list
                callbacks = list(self.__error_callbacks)

            # Call each error callback
            for callback, context in callbacks:
                callback(error, context=context)


    def attach_callback(self, callback, on_reception=True, on_transmission=True,
                        packet: Callable[[Packet], bool]=lambda pkt:True):
        """
        Attach a new packet callback to current connector.

        :param callback: Processing function.
        :param on_reception: Boolean indicating if the callback monitors reception.
        :param on_transmission: Boolean indicating if the callback monitors transmission.
        :param filter: Lambda function filtering packets matching the callback.
        :returns: Boolean indicating if the callback has been successfully attached.
        """
        with self.__callbacks_lock:
            callbacks_dicts = (
                ([self.__reception_callbacks] if on_reception else []) +
                ([self.__transmission_callbacks] if on_transmission else [])
            )
            for callback_dict in callbacks_dicts:
                callback_dict[callback] = packet

        return len(callbacks_dicts) > 0

    def detach_callback(self, callback, on_reception=True, on_transmission=True):
        """
        Detach an existing packet callback from current connector.

        :param callback: Processing function.
        :param on_reception: Boolean indicating if the callback was monitoring reception.
        :param on_transmission: Boolean indicating if the callback was monitoring transmission.
        :returns: Boolean indicating if the callback has been successfully detached.
        """
        # Enter critical section
        with self.__callbacks_lock:
            removed = False
            callbacks_dicts = (
                ([self.__reception_callbacks] if on_reception else []) +
                ([self.__transmission_callbacks] if on_transmission else [])
            )
            for callback_dict in callbacks_dicts:
                if callback in callback_dict:
                    del callback_dict[callback]
                    removed = True

        return removed

    def migrate_callbacks(self, connector):
        """Migrate callbacks to another connector
        """
        # Enter critical section
        with self.__callbacks_lock:
            for cb, cb_filter in self.__reception_callbacks.items():
                connector.attach_callback(
                    cb,
                    packet=cb_filter,
                    on_reception=True
                )
            self.__reception_callbacks = {}

            for cb, cb_filter in self.__transmission_callbacks.items():
                connector.attach_callback(
                    cb, packet=cb_filter, on_transmission=True
                )
            self.__transmission_callbacks = {}


    def reset_callbacks(self, reception = True, transmission = True):
        """
        Detach any packet callback attached to the current connector.

        :param on_reception: Boolean indicating if the callbacks monitoring
                             reception are detached.
        :param on_transmission: Boolean indicating if the callbacks monitoring
                                transmission are detached.
        """

        # Remove all callbacks
        with self.__callbacks_lock:
            if reception:
                self.__reception_callbacks = {}
            if transmission:
                self.__transmission_callbacks = {}


    def monitor_packet_tx(self, packet):
        """
        Signals the transmission of a packet and triggers execution of matching
        transmission callbacks.

        :param packet: scapy packet being transmitted from whad-client.
        """
        # Enter critical section
        with self.__callbacks_lock:
            for callback,packet_filter in self.__transmission_callbacks.items():
                if packet_filter(packet):
                    callback(packet)

    def monitor_packet_rx(self, packet):
        """
        Signals the reception of a packet and triggers execution of matching reception callbacks.

        :param packet: scapy packet being received by whad-client.
        """
        # Enter critical section
        with self.__callbacks_lock:
            for callback,packet_filter in self.__reception_callbacks.items():
                if packet_filter(packet):
                    callback(packet)

    def set_device(self, device=None):
        """
        Set device linked to this connector.

        :param WhadDevice device: Device to be used with this connector.
        """
        if device is not None:
            self.__device = device

    def mark_stalled(self):
        """Mark connector as stalled (pending disconnection)
        """
        self.__stalled = True


    def is_stalled(self) -> bool:
        """Determine if the interface associated with this connector is stalled,
        i.e. has messages awaiting processing even if closed.

        :return: True if interface is stalled, False otherwise.
        :rtype: bool
        """
        return self.__stalled

    @property
    def device(self):
        """Get the connector associated device instance
        """
        return self.__device

    @property
    def hub(self) -> ProtocolHub:
        """Get the connector protocol hub

        :return: Instance of ProtocolHub
        :rtype: ProtocolHub
        """
        return self.__device.hub

    def enable_synchronous(self, enabled : bool, events: bool = False):
        """Enable or disable synchronous mode

        Synchronous mode is a mode in which the connector expects sone third-party code to
        retrieve the received packets instead of forwarding them to the `on_packet()` callback.
        It is then possible to wait for some packet to be received and avoid the automatic
        behavior triggered by a call to `on_packet()`.

        :param enabled: If set to `True`, enable synchronous mode. Otherwise disable it.
        :type enabled: bool
        :param events: If set to `True`, synchronous mode will also capture events
                       sent by the associate device
        :type events: bool, optional
        """
        # Clear pending packets if we are disabling this feature.
        if not enabled and self.__sync_mode != Connector.SYNC_MODE_OFF:
            self.__sync_mode = Connector.SYNC_MODE_OFF
        elif enabled:
            if events:
                self.__sync_mode = Connector.SYNC_MODE_ALL
            else:
                self.__sync_mode = Connector.SYNC_MODE_PKT

        # Clear events queue
        self.__sync_events.queue.clear()

    def is_synchronous(self):
        """Determine if the conncetor is in synchronous mode.

        :return: `True` if synchronous mode is enabled, `False` otherwise.
        """
        return self.__sync_mode != Connector.SYNC_MODE_OFF

    def add_sync_event(self, event: DeviceEvt):
        """Add an event to the synchronous event queue when synchronous mode is
        enabled.

        :param event: Device event to add to our queue of received events
        :type event: whad.device.DeviceEvt
        """
        # Insert device events in synchronous event queue only if SYNC_MODE_ALL
        if isinstance(event, DeviceEvt) and self.__sync_mode == Connector.SYNC_MODE_ALL:
            self.__sync_events.put(event)
        elif isinstance(event, MessageReceived) and self.__sync_mode >= Connector.SYNC_MODE_PKT:
            # If SYNC_MODE_PKT is enabled, add MessageReceived event into our 
            # synchronous events queue
            self.__sync_events.put(event)

    def wait_packet(self, timeout:float = None):
        '''Wait for a packet when in synchronous mode. This method should be only used
        with SYNC_MODE_PKT to avoid discarding any device event.

        :param timeout: If specified, defines a timeout when querying the PDU queue
        :type timeout: float, optional
        :return: Received packet if any, None if empty or when timeout is reached
        :rtype: scapy.packet.Packet
        '''
        if self.__sync_mode >= Connector.SYNC_MODE_PKT:
            try:
                event = self.__sync_events.get(block=True, timeout=timeout)
                if isinstance(event, MessageReceived):
                    pkt = event.message.to_packet()
                    if pkt is not None:
                        return pkt
            except Empty:
                return None
        else:
            return None

    def lock(self):
        """Lock connector. A locked connector will not dispatch packets/pdus like
        in synchronous mode and will keep them in a waiting queue, but will dispatch
        them all at once when unlocked.
        """
        logger.info("[connector] lock()")
        self.__locked = True

        # Clear pending PDUs queue
        with self.__locked_pdus.mutex:
            self.__locked_pdus.queue.clear()

    def unlock(self, dispatch_callback=None):
        """Unlock connector and dispatch pending PDUs.

        :param  dispatch_callback: PDU dispatch callback that overrides the
                                   internal dispatch routine
        :type   dispatch_callback: callable
        """
        with self.__lock:
            logger.info("[connector][%s] unlock()", self.device.interface)
            # Dispatch PDUs
            try:
                # Loop until locked PDUs queue is empty
                while not self.__locked_pdus.empty():
                    # Retrieve PDU
                    message = self.__locked_pdus.get(block=True, timeout=0.2)
                    logger.debug("[connector][%s] Unlocked message for processing: %s",
                                self.device.interface, message)
                    if dispatch_callback is None:
                        logger.debug("[connector][%s] forward to __process_pkt_message()")
                        self.__process_pkt_message(message)
                    else:
                        # Call the provided dispatch callback
                        dispatch_callback(message)

                    # Mark locked PDU as processed
                    self.__locked_pdus.task_done()
            except Empty:
                logger.debug("[connector][%s] Processed all messages", self.device.interface)

        # Mark connector as unlocked
        self.__locked = False

    def is_locked(self) -> bool:
        """Determine if the connector is locked.

        :return: `True` if lock mode is enabled, `False` otherwise.
        """
        logger.info("[connector][%s] is_locked() -> %s", self.device.interface, self.__locked)
        return self.__locked

    def add_locked_pdu(self, pdu):
        """Add a pending Protocol Data Unit (PDU) to our locked pdus queue.

        :param  Packet pdu:  Packet to add to locked packets queue
        :type   pdu: scapy.packet.Packet
        """
        # We use the same lock used when unlocking to prevent adding more locked PDUs into
        # our locked PDUs queue, until it becomes full and connector is unlocked.
        with self.__lock:
            logger.info("[connector][%s] Add locked pdu: %s", self.device.interface, pdu)
            self.__locked_pdus.put(pdu)

    def has_locked_pdus(self) -> bool:
        """Determine if connector has locked PDUs.

        :return: `True` if connector has locked PDUs, `False` otherwise.
        :rtype: bool
        """
        return not self.__locked_pdus.empty()

    # Device interaction
    def send_message(self, message, keep=None):
        """Sends a message to the underlying device without waiting for an answer.

        :param Message message: WHAD message to send to the device.
        :param filter: optional filter function for incoming message queue.
        """
        try:
            logger.debug("sending WHAD message to device: %s", message)
            self.__device.send_message(message, keep)
        except WhadDeviceError as device_error:
            logger.debug("an error occurred while communicating with the WHAD device !")
            self.on_error(device_error)

    def send_command(self, message, keep=None):
        """Sends a command message to the underlying device and waits for an answer.

        By default, this method will wait for a CmdResult message, but you can provide
        any other filtering function/lambda if you are expecting another message as a
        reply from the device.

        :param Message message: WHAD message to send to the device
        :param filter: Filtering function used to match the expected response from the device.
        """
        try:
            return self.__device.send_command(message, keep=keep)
        except WhadDeviceError as device_error:
            logger.debug("an error occurred while communicating with the WHAD device !")
            self.on_error(device_error)
            return None

    ######################################
    # Packet flow handling
    ######################################

    def send_packet(self, packet):
        """Send packet to our device.
        """
        # Monitor this outgoing packet
        self.monitor_packet_tx(packet)

        # Convert packet into the corresponding message
        msg = self.hub.convert_packet(packet)

        if msg is not None:
            logger.info("[connector] send packet command")
            resp = self.send_command(msg, message_filter(CommandResult))
            logger.info("[connector] Command sent, result: %s", resp)

            # Do we have an error while sending this command ?
            if resp is None:
                # Report WHAD device as disconnected
                raise WhadDeviceDisconnected()

            # Check if command was successful
            return isinstance(resp, Success)

        # Cannot convert packet
        logger.error(("[connector] Packet cannot be converted into the"
                        "corresponding WHAD message"))
        return False


    def wait_for_message(self, timeout=None, keep=None, command=False):
        """Waits for a specific message to be received.

        This method reads the message queue and return the first message that matches the
        provided filter. A timeout can be specified and will cause this method to return
        None if this timeout is reached.
        """
        if self.__device is not None:
            return self.__device.wait_for_message(keep=keep, timeout=timeout, command=command)

    # Message callbacks
    def on_any_msg(self, message): # pylint: disable=W0613
        """Callback function to process any incoming messages.

        This method MAY be overriden by inherited classes.

        :param message: WHAD message
        """

    def on_discovery_msg(self, message): # pylint: disable=W0613
        """Callback function to process incoming discovery messages.

        This method MUST be overriden by inherited classes.

        :param message: Discovery message
        """
        logger.error("method `on_discovery_msg` must be implemented in inherited classes")
        raise RequiredImplementation()

    def on_generic_msg(self, message): # pylint: disable=W0613
        """Callback function to process incoming generic messages.

        This method MUST be overriden by inherited classes.

        :param message: Generic message
        """
        logger.error("method `on_generic_msg` must be implemented in inherited classes")
        raise RequiredImplementation()

    def on_domain_msg(self, domain, message): # pylint: disable=W0613
        """Callback function to process incoming domain-related messages.

        This method MUST be overriden by inherited classes.

        :param message: Domain message
        """
        logger.error("method `on_domain_msg` must be implemented in inherited classes")
        raise RequiredImplementation()


    def on_packet(self, packet): # pylint: disable=W0613
        """Callback function to process incoming packets.

        This method MUST be overriden by inherited classes.

        :param packet: Packet
        :type packet: :class:`scapy.packet.Packet`
        """
        logger.error("method `on_packet` must be implemented in inherited classes")
        raise RequiredImplementation()

    # TODO: event for device and connector event ?
    def on_event(self, event): # pylint: disable=W0613
        """Callback function to process incoming events.

        This method MUST be overriden by inherited classes.

        :param event: Event to process
        :type event: :class:`whad.hub.events.AbstractEvent`
        """
        logger.error("Class: %s", self.__class__)
        logger.error("method `on_event` must be implemented in inherited classes")
        raise RequiredImplementation()


    def send_event(self, event: DeviceEvt):
        """Send an event into the connector event queue.

        :param event: Event to add to the connector's event queue
        :type event: DeviceEvt
        """
        self.__events.put(event)

    def add_listener(self, listener: Callable[..., None],event_cls: Union[List[Event],
                                                                    Event] = None):
        """Add a connector notification listener with optional event filter.

        :param listener: callable to handle events
        :type listener: callable
        :param event_cls: List of event classes or single event class to match
        :type event_cls: list, ConnectorEvent, optional
        """
        if event_cls is not None:
            events = event_cls if isinstance(event_cls, list) else [event_cls]
        else:
            events = None
        self.__listeners.append((listener, events))

    def remove_listener(self, listener: Callable[..., None]) -> bool:
        """Remove listener from registered listeners.
        """
        # Find one or more entries related to the given listener
        items = set(filter(lambda x: x[0] == listener, self.__listeners))

        # Nothing found, return False
        if len(items) == 0:
            return False

        # Remove found entries.
        for item in items:
            self.__listeners.remove(item)

        # Success
        return True

    def clear_listeners(self):
        """Clear listeners.
        """
        self.__listeners = []

    def notify(self, event):
        """Notify listeners of a specific event.
        """
        for listener, events in self.__listeners:
            if events is None:
                listener(event)
            elif event.__class__ in events:
                listener(event)

    @contextlib.contextmanager
    def get_event(self, timeout: Optional[float] = None) -> Generator[DeviceEvt, None, None]:
        """Retrieve event from connector's event queue.

        :param timeout: Timeout in seconds
        :type timeout: float
        """
        try:
            yield self.__events.get(timeout=timeout)
        except Empty as err:
            raise err

        self.__events.task_done()

    def busy(self) -> bool:
        """Determine if this connector is busy.
        """
        # In synchronous mode, a busy connector has unprocessed events to
        # left in its events queue. If this events queue is empty, it is not considered
        # busy anymore.
        if self.__sync_mode != Connector.SYNC_MODE_OFF:
            return not self.__sync_events.empty()

        # If connector is locked or has unprocessed locked PDUs, it is considered busy.
        if not self.__locked_pdus.empty():
            return True

        # If not in synchronous mode, connector is busy if it still has events to
        # process (incoming messages) or if the associated interface has
        # messages to send
        return not self.__events.empty() or self.device.busy()

    def on_disconnection(self):
        """Device has disconnected or been closed.
        """
        logger.debug("[%s] interface has disconnected", self.device.interface)

        # Mark connector as disconnected
        self.__disconnected.set()

    def on_device_event(self, event: DeviceEvt):
        """Dispatch message to the connector's handlers.

        This method may trigger specific message processing in inherited
        connector's classes as well as attached protocol stacks. Since it
        is only called by the connector's I/O thread, that's pretty safe.

        :param event: Device event to process
        :type event: whad.device.DeviceEvt
        """
        # If synchronous mode is enabled and capturing all events, we simply
        # add these events to the synchronous mode event queue.
        if self.__sync_mode == Connector.SYNC_MODE_ALL:
            logger.debug("[synchronous] received event %s and save into sniffing queue", event)
            self.add_sync_event(event)
        else:
            # Did we receive a disconnection event ?
            if isinstance(event, Disconnected):
                logger.debug("[%s] received a disconnection message, processing ...",
                             self.device.interface)
                # Notify disconnection if we are not locked
                if self.is_locked():
                    logger.debug("[%s] locked, mark connector as stalled", self.device.interface)
                    self.mark_stalled()
                else:
                    logger.debug("[%s] not stalled, report disconnection", self.device.interface)
                    self.on_disconnection()

            # Or a hub emssage ?
            elif isinstance(event, MessageReceived):
                # If synchronous mode is enabled, add events into our synchronous
                # event queue. At this point in code, we are certain to be in
                # SYNC_MODE_PKT mode.
                if self.__sync_mode == Connector.SYNC_MODE_PKT:
                    self.add_sync_event(event)
                else:
                    # Process hub message if not in synchronous mode
                    self.process_message(event.message)

    # pylint: disable=C0301
    def sniff(self, messages: List = None, timeout: float = None) -> Generator[HubMessage, None, None]:
        """Enable sniffing mode and report any received messages, optionally
        filtered by their type/classes if `messages` is provided.

        :param messages: If specified, sniff only messages that match the given types.
        :param messages: List, optional
        :param timeout: If specified, set a sniffing timeout in seconds
        :type timeout: float, optional
        """
        # Enable sniffing mode (and disable message processing)
        self.enable_synchronous(True, events=True)

        # Listen for messages
        initial_to = timeout
        start_mark = time()
        while True:
            try:
                # Wait for an event
                event = self.__sync_events.get(block=True, timeout=timeout)
                logger.debug("[sniffer][%s] received event %s",
                                     self.device.interface, event)

                # If we received an event, process it
                if isinstance(event, Disconnected):
                    # Interface has disconnected, log this error and exit the function
                    logger.debug("[sniffer][%s] Interface has disconnected !",
                                    self.device.interface)
                    # Maybe raise an exception instead of simply exiting the sniff() method ?
                    # Calling code cannot tell if sniffing is just done or if an error occurred.
                    return

                if isinstance(event, MessageReceived):
                    logger.debug("[sniffer][%s] received message, processing",
                                     self.device.interface)
                    # Retrieve message
                    message = event.message

                    # Do we need to filter messages by type ?
                    if messages is not None:
                        logger.debug(
                            "[sniffer][%s] checking message type (%s) against filtered types (%s)",
                            self.device.interface, type(message), messages
                        )
                        if isinstance(message, messages):
                            yield message
                    else:
                        yield message

                # Notify queue we are done with this message
                self.__sync_events.task_done()

            except Empty:
                # We receive this exception when timeout has been reached when calling
                # our queue's get() method with a given timeout.
                logger.debug("[sniffer][%s] Sniffing timeout reached (%s seconds) !",
                             self.device.interface, initial_to)

            # If timeout is provided, compute remaining time and
            # exit loop when reached.
            if timeout is not None:
                timeout = timeout - (time() - start_mark)
                if timeout < 0:
                    break

        # Sniffing done, disable sniffing mode and return to normal operation
        self.enable_synchronous(False)

    def process_message(self, message: HubMessage):
        """Process received message.
        """
        logger.debug("[connector][%s] process_message() called for message %s", self.device.interface,
                     message)
        # Forward message to the on_any_msg() handler
        self.on_any_msg(message)

        # If message is of type "discovery", forward to our discovery
        # handler
        if message.message_type == "discovery":
            logger.info("message is about device discovery, forwarding to discovery handler")
            self.on_discovery_msg(message)
        elif message.message_type == "generic":
            # Handle generic result message
            if isinstance(message, CommandResult):
                if message.result_code == CommandResult.UNSUPPORTED_DOMAIN:
                    logger.error("domain not supported by this device")
                    raise UnsupportedDomain("")

            # Forward to generic message handler
            logger.info("message is generic, forwarding to default handler")
            self.on_generic_msg(message)
        else:
            domain = message.message_type
            if domain is not None:
                # Check if message is a received packet
                if issubclass(message, AbstractPacket):
                    # If connector is locked, save message into locked pdus
                    if self.is_locked():
                        self.add_locked_pdu(message)
                    else:
                        self.__process_pkt_message(message)
                # Check if message is a received event
                elif issubclass(message, AbstractEvent):
                    # Convert message into event
                    event = message.to_event()
                    if event is not None:
                        # Forward to our connector
                        self.on_event(event)

                # Other messages go to on_domain_msg
                else:
                    logger.info("message concerns domain `%s`, forward to domain-specific handler",
                                domain)
                    self.on_domain_msg(domain, message)

    def __process_pkt_message(self, message: HubMessage):
        """Process a Hub message containing a packet.

        :param message: Message supporting the AbstractPacket interface
        :type message: HubMessage
        """
        # Convert message into packet
        packet = message.to_packet()
        if packet is not None:
            self.monitor_packet_rx(packet)
            self.on_packet(packet)

    def join(self):
        """Wait for the interface to disconnect and messages to be processed.
        """
        self.__disconnected.wait()


class LockedConnector(Connector):
    """Provides a lockable connector.
    """

    def __init__(self, device):
        # We set the connector with no interface for now
        super().__init__(None)

        # Then we lock it
        self.lock()

        # And we eventually configure the interface
        # Once the device connector is set, packets will go in a locked queue
        # and could be later retrieved when connector is unlocked.
        self.set_device(device)
        device.set_connector(self)


class WhadDeviceConnector(Connector):
    """
    This class is an alias for :py:class:`whad.device.connector.Connector`,
    and will be deprecated in a near future. This class has been introduced
    in a previous version of WHAD and has been renamed for clarity purpose.
    """



from whad.dot15d4.stack.manager import Dot15d4Manager
from whad.dot15d4.stack.service import Dot15d4Service
from whad.dot15d4.stack.mac.constants import MACAddressMode
from whad.dot15d4.stack.mac.helpers import is_short_address

from whad.zigbee.stack.aps.exceptions import APSTimeoutException
from whad.zigbee.stack.aps.database import APSIB
from whad.zigbee.stack.aps.security import APSApplicationLinkKeyData, \
    APSNetworkKeyData, APSTrustCenterLinkKeyData
from whad.zigbee.stack.aps.constants import APSSecurityStatus, \
    APSSourceAddressMode, APSDestinationAddressMode, APSKeyType
from whad.zigbee.stack.nwk.constants import NWKAddressMode
from whad.zigbee.crypto import ApplicationSubLayerCryptoManager

from whad.zigbee.stack.apl import APLManager
from whad.common.stack import alias, source, state

from scapy.layers.zigbee import ZigbeeSecurityHeader, ZigbeeAppDataPayload, \
    ZigbeeAppCommandPayload, ZigbeeClusterLibrary, ZigbeeDeviceProfile

from whad.exceptions import RequiredImplementation

import logging

logger = logging.getLogger(__name__)


class APSService(Dot15d4Service):
    """
    This class represents an APS service, exposing a standardized API.
    """
    def __init__(self, manager, name=None):
        super().__init__(manager, name=name, timeout_exception_class=APSTimeoutException)


class APSDataService(APSService):
    """
    APS service processing Data frames.
    """
    def __init__(self, manager):
        super().__init__(manager, name="aps_data")

    @Dot15d4Service.request("APSDE-DATA")
    def data(
                self,
                asdu,
                destination_address_mode,
                destination_address,
                destination_endpoint,
                profile_id,
                cluster_id,
                source_endpoint,
                alias_address=None,
                alias_sequence_number=0,
                radius=30,
                security_enabled_transmission=False,
                use_network_key=False,
                acknowledged_transmission=False,
                fragmentation_permitted=False,
                include_extended_nonce=False
    ):
        """
        Implements the APSDE-DATA Request.

        Processes and transmits Data PDU to lowest layers.
        """
        multicast = False
        # If transmission must be secured, select the key pair, encrypt and add a security header accordingly
        if security_enabled_transmission:
            apsDeviceKeyPairSet = self.database.get("apsDeviceKeyPairSet")
            candidate_keys = apsDeviceKeyPairSet.select(destination_address, unverified=False)
            if len(candidate_keys) == 0:
                return False
            candidate_key = candidate_keys[0]
            key_identifier = 0
            asdu = ZigbeeSecurityHeader(
                key_type=key_identifier,
                extended_nonce=int(include_extended_nonce),
                fc=candidate_key.outgoing_frame_counter,
                data=bytes(asdu)
            )
            if asdu.extended_nonce:
                asdu.source=self.manager.get_layer('nwk').database.get("nwkIeeeAddress")

            crypto_manager = ApplicationSubLayerCryptoManager(candidate_key.key, None)
            asdu = crypto_manager.encrypt(asdu)


        if destination_address_mode == APSDestinationAddressMode.DST_ADDRESS_AND_DST_ENDPOINT_NOT_PRESENT:
            raise RequiredImplementation("BindingTableSearch")

        elif destination_address_mode == APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT:
            apdu = ZigbeeAppDataPayload(
                delivery_mode = 0,
                dst_endpoint = destination_endpoint,
                src_endpoint = source_endpoint,
                profile=profile_id,
                cluster=cluster_id,
                counter=self.manager.database.get("apsCounter")
            )/asdu

        elif destination_address_mode == APSDestinationAddressMode.EXTENDED_ADDRESS_DST_ENDPOINT_PRESENT:
            nwkAddressMap = self.manager.get_layer('nwk').database.get("nwkAddressMap")
            short_destination_address = None
            if destination_address in nwkAddressMap:
                destination_address = nwkAddressMap[destination_address]
            else:
                return False
            apdu = ZigbeeAppDataPayload(
                dst_endpoint=destination_endpoint,
                cluster=cluster_id,
                profile=profile_id,
                src_endpoint=source_endpoint,
                counter=self.manager.database.get("apsCounter")
            )/asdu
        elif destination_address_mode == APSDestinationAddressMode.SHORT_GROUP_ADDRESS_DST_ENDPOINT_NOT_PRESENT:
            apdu = ZigbeeAppDataPayload(
                delivery_mode = 3,
                group_addr = destination_address,
                src_endpoint = source_endpoint,
                profile=profile_id,
                cluster=cluster_id,
                counter=self.manager.database.get("apsCounter")
            )/asdu
            if self.manager.database.get("nwkUseMulticast"):
                multicast = True
                apdu = ZigbeeAppDataPayload(
                    delivery_mode = 2,
                    dst_endpoint = 0xFF,
                    src_endpoint = source_endpoint,
                    profile=profile_id,
                    cluster=cluster_id,
                    counter=self.manager.database.get("apsCounter")
                )/asdu

            else:
                multicast = False
                destination_address = 0xFFFD
                apdu = ZigbeeAppDataPayload(
                    delivery_mode = 3,
                    group_addr = destination_address,
                    src_endpoint = source_endpoint,
                    profile=profile_id,
                    cluster=cluster_id,
                    counter=self.manager.database.get("apsCounter")
                )/asdu

        if alias_address is not None and acknowledged_transmission:
            return False

        if acknowledged_transmission:
            apdu.frame_control.ack_req = True
            raise RequiredImplementation("APSAckImplementation")
        else:
            apdu.frame_control.ack_req = False

        if security_enabled_transmission:
            apdu.frame_control.security = True
        else:
            apdu.frame_control.security = False

        counter = self.manager.database.get("apsCounter")
        self.manager.database.set("apsCounter", counter+1)
        return self.manager.get_layer('nwk').get_service("data").data(
            apdu,
            nsdu_handle=0,
            alias_address=alias_address,
            alias_sequence_number=alias_sequence_number,
            destination_address_mode=NWKAddressMode.MULTICAST if multicast else NWKAddressMode.UNICAST,
            destination_address=destination_address,
            radius=radius,
            non_member_radius=self.database.get("apsNonmemberRadius"),
            discover_route=False,
            security_enable=use_network_key
        )

    @Dot15d4Service.indication("APSDE-DATA")
    def indicate_data(
                        self,
                        apdu,
                        destination_address_mode,
                        destination_address,
                        source_address,
                        security_status,
                        link_quality
    ):
        """
        Implements the APSDE-DATA Indication.

        Processes and transmits an ASDU to upper layer.
        """
        if ZigbeeSecurityHeader in apdu:
            asdu = ZigbeeAppDataPayload(apdu.data)
        else:
            asdu = apdu

        if 'extended_hdr' in asdu.frame_control and asdu.aps_frametype in (0, 2):
            raise RequiredImplementation("Fragmentation")

        if hasattr(asdu, "dst_endpoint"):
            dst_endpoint = asdu.dst_endpoint
        else:
            dst_endpoint = None

        if hasattr(asdu, "src_endpoint"):
            src_endpoint = asdu.src_endpoint
        else:
            src_endpoint = None

        if hasattr(asdu, "profile"):
            profile_id = asdu.profile
        else:
            profile_id = None

        if hasattr(asdu, "cluster"):
            cluster_id = asdu.cluster
        else:
            cluster_id = None

        nwkAddressMap = self.manager.get_layer('nwk').database.get("nwkAddressMap")
        selected_extended_address = None
        '''
        for extended_address, short_address in nwkAddressMap.items():
            if short_address == source_address:
                selected_extended_address = extended_address
                break
        '''
        source = None
        source_address_mode = None

        if selected_extended_address is None:
            source = source_address
            source_address_mode = APSSourceAddressMode.SHORT_ADDRESS_SRC_ENDPOINT_PRESENT
        else:
            source = selected_extended_address
            if src_endpoint is None:
                source_address_mode = APSSourceAddressMode.EXTENDED_ADDRESS_SRC_ENDPOINT_NOT_PRESENT
            else:
                source_address_mode = APSSourceAddressMode.EXTENDED_ADDRESS_SRC_ENDPOINT_PRESENT

        destination = destination_address
        if destination_address_mode == NWKAddressMode.MULTICAST:
            destination_mode = APSDestinationAddressMode.SHORT_GROUP_ADDRESS_DST_ENDPOINT_NOT_PRESENT
        else:
            if is_short_address(destination_address):
                destination_mode = APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT
            else:
                if dst_endpoint is not None:
                    destination_mode = APSDestinationAddressMode.EXTENDED_ADDRESS_DST_ENDPOINT_PRESENT
                else:
                    destination_mode = APSDestinationAddressMode.EXTENDED_ADDRESS_DST_ENDPOINT_NOT_PRESENT

        if profile_id == 0x0000:
            payload = asdu[ZigbeeDeviceProfile]
        else:
            payload = asdu[ZigbeeClusterLibrary]

        if 'ack_req' in asdu.frame_control:
            acknowledgement = ZigbeeAppDataPayload(
                frame_control = 0,
                delivery_mode = 0,
                aps_frametype = 2,
                dst_endpoint = src_endpoint,
                src_endpoint = dst_endpoint,
                profile=profile_id,
                cluster=cluster_id,
                counter=asdu.counter
            )

            self.manager.get_layer('nwk').get_service("data").data(
                acknowledgement,
                nsdu_handle=0,
                destination_address_mode=NWKAddressMode.UNICAST,
                destination_address=source_address,
                radius=30,
                non_member_radius=self.database.get("apsNonmemberRadius"),
                discover_route=False,
                security_enable=True
            )

        return (
                    payload,
                    {
                        "destination_address":destination,
                        "destination_address_mode":destination_mode,
                        "destination_endpoint":dst_endpoint,
                        "source_address":source,
                        "source_address_mode":source_address_mode,
                        "source_endpoint":src_endpoint,
                        "profile_id":profile_id,
                        "cluster_id":cluster_id,
                        "security_status":security_status,
                        "link_quality":link_quality
                    }
        )

    def on_data_apdu(
                        self,
                        apdu,
                        destination_address_mode,
                        destination_address,
                        source_address,
                        security_status,
                        link_quality
    ):
        """
        Callback processing APDU data forwarded by the APS Manager.
        """
        self.indicate_data(
                            apdu,
                            destination_address_mode,
                            destination_address,
                            source_address,
                            security_status,
                            link_quality
        )

class APSManagementService(APSService):
    """
    APS service processing Management operations.
    """
    def __init__(self, manager):
        super().__init__(manager, name="aps_management")


    @Dot15d4Service.request("APSME-GET")
    def get(self, attribute):
        """
        Implement the APSME-GET request operation.

        Gets an attribute value from the APSIB database.
        """
        return self.database.get(attribute)

    @Dot15d4Service.request("APSME-SET")
    def set(self, attribute, value):
        """
        Implement the APSME-SET request operation.

        Sets an attribute value from the APSIB database.
        """
        return self.database.set(attribute, value)

    @Dot15d4Service.indication("APSME-TRANSPORT-KEY")
    def indicate_transport_key(self, source_address, standard_key_type, transport_key_data):
        """
        Implement the APSME-TRANSPORT-KEY indication operation.

        Indicates to upper layer that a transport key has been received.
        """
        return (
            transport_key_data,
            {
                "standard_key_type":standard_key_type,
                "source_address":source_address
            }
        )

    @Dot15d4Service.request("APSME-TRANSPORT-KEY")
    def transport_key(self, destination_address, standard_key_type, transport_key_data):

        if standard_key_type == APSKeyType.TRUST_CENTER_LINK_KEY:
            raise RequiredImplementation("TrustCenterLinkKey")
        elif standard_key_type == APSKeyType.STANDARD_NETWORK_KEY:
            nwkAddressMap = self.manager.get_layer('nwk').database.get("nwkAddressMap")
            selected_destination_address = None
            for extended_address, short_address in nwkAddressMap.items():
                if short_address == destination_address:
                    selected_destination_address = extended_address
                    break

            apdu = ZigbeeAppDataPayload(
                delivery_mode=0,
                frame_control=['security'],
                aps_frametype=1
            )
            asdu = ZigbeeAppCommandPayload(
                cmd_identifier = 5,
                key_type = 1,
                key = transport_key_data.key,
                key_seqnum = transport_key_data.key_sequence_number,
                dest_addr = selected_destination_address,
                src_addr = self.manager.get_layer('nwk').database.get("nwkIeeeAddress")
            )
            #print(self.manager.get_layer('nwk').database.get("nwkIeeeAddress"))
            #asdu.show()

            apsDeviceKeyPairSet = self.database.get("apsDeviceKeyPairSet")
            candidate_keys = apsDeviceKeyPairSet.select(destination_address, unverified=False)

            if len(candidate_keys) == 0:
                return False
            candidate_key = candidate_keys[0]

            asdu = apdu / ZigbeeSecurityHeader(
                key_type=2,
                extended_nonce=1,
                fc=candidate_key.outgoing_frame_counter,
                nwk_seclevel=5,
                data=bytes(asdu)
            )


            asdu.source=self.manager.get_layer('nwk').database.get("nwkIeeeAddress")
            crypto_manager = ApplicationSubLayerCryptoManager(candidate_key.key, 0x00)
            asdu = crypto_manager.encrypt(asdu)

            return self.manager.get_layer('nwk').get_service("data").data(
                asdu,
                nsdu_handle=0,
                destination_address_mode=NWKAddressMode.UNICAST,
                destination_address=(
                    transport_key_data.parent_address if
                    transport_key_data.use_parent else
                    destination_address
                ),
                discover_route=False,
                security_enable=False
            )
        elif standard_key_type == APSKeyType.APPLICATION_LINK_KEY:
            apdu = ZigbeeAppDataPayload(
                delivery_mode=0,
            ) / ZigbeeAppCommandPayload(
                cmd_identifier = 5,
                key_type = standard_key_type,
                key = transport_key_data.key,
                partner_addr = transport_key_data.partner_address,
                dest_addr = destination_address,
                initiator = 1,
                src_addr = self.manager.get_layer('nwk').database.get("nwkNetworkAddress")
            )
            return self.manager.get_layer('nwk').get_service("data").data(
                apdu,
                nsdu_handle=0,
                destination_address_mode=NWKAddressMode.UNICAST,
                destination_address=(
                    transport_key_data.parent_address if
                    transport_key_data.use_parent else
                    destination_address
                ),
                discover_route=False,
                security_enable=False
            )

    def process_transport_key(self, nsdu, source_address, security_status):
        """
        Extract the transport key from the provided NSDU and process it.
        """
        # Check if the key matches the security policy
        authorized = self.database.get("apsTrustCenterAddress") is not None
        if ZigbeeSecurityHeader in nsdu:
            asdu = ZigbeeAppCommandPayload(nsdu.data)
        else:
            asdu = nsdu[ZigbeeAppCommandPayload]

        if (
                (
                    asdu.key_type in (APSKeyType.APPLICATION_LINK_KEY, APSKeyType.TRUST_CENTER_LINK_KEY) and
                    authorized and security_status != APSSecurityStatus.SECURED_LINK_KEY
                ) or
                (asdu.key_type == 1 and security_status != APSSecurityStatus.SECURED_LINK_KEY)
            ):
            logger.info("[aps_management] transport key doesn't match the security policy, discarding.")
            return

        # If we got an APS Application Link Key, indicate transport key using APSME-TRANSPORT-KEY
        if asdu.key_type == APSKeyType.APPLICATION_LINK_KEY:
            source = source_address
            standard_key_type = asdu.key_type
            transport_key_data = APSApplicationLinkKeyData(asdu.key, asdu.src_addr)

            return self.indicate_transport_key(source, standard_key_type, transport_key_data)

        # If we got a Trust Center Link Key (1) or a Network Key (2) AND it is transmitted to us,
        # indicate key to upper layer using APSME-TRANSPORT-KEY indication.
        if (
                (
                    asdu.key_type in (APSKeyType.STANDARD_NETWORK_KEY, APSKeyType.TRUST_CENTER_LINK_KEY) and
                    asdu.dest_addr == self.manager.get_layer('nwk').database.get("nwkIeeeAddress")
                )
                or
                ( asdu.key_type == APSKeyType.STANDARD_NETWORK_KEY and asdu.dest_addr == 0x0000000000000000 )
        ):

            source = asdu.src_addr
            standard_key_type = asdu.key_type

            if asdu.key_type == APSKeyType.STANDARD_NETWORK_KEY:
                transport_key_data = APSNetworkKeyData(asdu.key, asdu.key_seqnum, False)
            else:
                transport_key_data = APSTrustCenterLinkKeyData(asdu.key)

            return self.indicate_transport_key(source, standard_key_type, transport_key_data)

        # If we got a Network key and dest address equals to 0xFFFFFFFFFFFFFFFF (distributed security network),
        # update APS Trust Center Address and indicate to upper layer using APSME-TRANSPORT-KEY indication.
        if asdu.key_type == APSKeyType.STANDARD_NETWORK_KEY and asdu.dest_addr == 0xFFFFFFFFFFFFFFFF:

            source = asdu.src_addr
            standard_key_type = asdu.key_type
            transport_key_data = APSNetworkKeyData(asdu.key, asdu.key_seqnum, False)
            self.database.set("apsTrustCenterAddress", 0xFFFFFFFFFFFFFFFF)

            return self.indicate_transport_key(source, standard_key_type, transport_key_data)

        # If the packet is not indicated for us, let's just route it to the destination using data service.
        if (
            asdu.key_type in (APSKeyType.STANDARD_NETWORK_KEY,APSKeyType.TRUST_CENTER_LINK_KEY) and
            asdu.dest_addr != self.manager.get_layer('nwk').database.get("nwkIeeeAddress")
        ):

            # Forward to destination device
            self.manager.get_layer('nwk').get_service("data").data(
                asdu,
                destination_address=asdu.dest_addr,
                security_enable=False
            )


    def on_command_apdu(
                        self,
                        nsdu,
                        destination_address_mode,
                        destination_address,
                        source_address,
                        security_status,
                        link_quality
    ):
        """
        Callback processing APDU commands transmitted by the APS Manager.
        """
        # Format APDU command to facilitate parsing
        if ZigbeeSecurityHeader in nsdu:
            asdu = ZigbeeAppCommandPayload(nsdu.data)
        else:
            asdu = nsdu[ZigbeeAppCommandPayload]

        # Forwards command to the right handler

        if asdu.cmd_identifier == 5: # APS_CMD_TRANSPORT_KEY
            self.process_transport_key(nsdu, source_address, security_status)
        # Some processing for other commands is missing here

    @Dot15d4Service.indication("APSME-JOIN")
    def indicate_join(
                        self,
                        network_address,
                        extended_address,
                        capability_information,
                        rejoin=False,
                        secure_rejoin=False
    ):
        return (network_address,
            {
                "extended_address":extended_address,
                "capability_information":capability_information,
                "rejoin":rejoin,
                "secure_rejoin":secure_rejoin
            }
        )

class APSInterpanPseudoService(APSService):
    """
    APS pseudo service forwarding InterPAN operations.

    This service is only there to forward InterPAN to upper layers.
    """
    def __init__(self, manager):
        super().__init__(manager, name="aps_interpan")

    @Dot15d4Service.request("INTRP-DATA")
    def interpan_data(
                        self,
                        asdu,
                        asdu_handle=0,
                        source_address_mode=MACAddressMode.SHORT,
                        destination_address_mode=MACAddressMode.SHORT,
                        destination_pan_id=0xFFFF,
                        destination_address=0xFFFF,
                        profile_id=0,
                        cluster_id=0,
                        acknowledged_transmission=False
    ):
        """
        Implements the INTRP-DATA Request.

        Transmits an InterPAN PDU using network layer.
        """
        return self.manager.get_layer('nwk').get_service("interpan").interpan_data(
            asdu,
            asdu_handle=asdu_handle,
            source_address_mode=source_address_mode,
            destination_pan_id=destination_pan_id,
            destination_address=destination_address,
            destination_address_mode=destination_address_mode,
            profile_id=profile_id,
            cluster_id=cluster_id,
            acknowledged_transmission=acknowledged_transmission
        )

    @Dot15d4Service.indication("INTRP-DATA")
    def indicate_interpan_data(
                                self,
                                asdu,
                                profile_id=0,
                                cluster_id=0,
                                destination_pan_id=0xFFFF,
                                destination_address=0xFFFF,
                                source_pan_id=0xFFFF,
                                source_address=0xFFFF,
                                link_quality=255
    ):
        """
        Implements INTRP-DATA indication.

        Indicates to upper layer that an InterPAN PDU has been received and processed by NWK & APS layers.
        """
        return (
            asdu,
            {
                "profile_id":profile_id,
                "cluster_id":cluster_id,
                "destination_pan_id":destination_pan_id,
                "destination_address":destination_address,
                "source_pan_id":source_pan_id,
                "source_address":source_address,
                "link_quality":link_quality
            }
        )

@state(APSIB)
@alias('aps')
class APSManager(Dot15d4Manager):
    """
    This class implements the Zigbee Application Support Sub-layer manager (APS).
    It provides a framework for application.

    It exposes two services providing the appropriate API + 1 pseudo service to forward InterPAN PDU.
    """
    def init(self):
        self.add_service("data", APSDataService(self))
        self.add_service("management", APSManagementService(self))
        self.add_service("interpan", APSInterpanPseudoService(self))


    def decrypt(self, pdu):
        """
        This method decrypts a PDU if the corresponding security material is found in the database.
        """
        # Check the presence of a security header
        if ZigbeeSecurityHeader not in pdu:
            logger.info("[aps] decryption failure - missing security header.")
            return pdu, False

        # Check if we have frame counter, source and key sequence number, mandatory for decryption
        if (
                not hasattr(pdu[ZigbeeSecurityHeader], "fc") or
                not hasattr(pdu[ZigbeeSecurityHeader], "source") or
                not hasattr(pdu[ZigbeeSecurityHeader], "key_seqnum")
        ):
            logger.info("[aps] decryption failure - missing mandatory data in security header.")
            return pdu, False

        # Extract mandatory informations from packet
        received_frame_count = pdu[ZigbeeSecurityHeader].fc
        key_sequence_number = pdu[ZigbeeSecurityHeader].key_seqnum
        sender_address = pdu[ZigbeeSecurityHeader].source
        key_identifier = pdu[ZigbeeSecurityHeader].key_type

        # Collect short address from network layer
        nwkAddressMap = self.get_layer('nwk').database.get("nwkAddressMap")
        short_address = None
        if sender_address in nwkAddressMap:
            short_address = nwkAddressMap[sender_address]

        # Extracty the KeyPair
        apsDeviceKeyPairSet = self.database.get("apsDeviceKeyPairSet")
        candidate_keys = apsDeviceKeyPairSet.select(short_address)
        for candidate_key in candidate_keys:
            if key_identifier == 0:
                input = None
            elif key_identifier == 2:
                input = 0x00
            elif key_identifier == 3:
                input = 0x02

            # Decryption process, implementation in ApplicationSubLayerCryptoManager
            crypto_manager = ApplicationSubLayerCryptoManager(candidate_key.key, input)
            decrypted, success = crypto_manager.decrypt(pdu)
            if success:
                return decrypted, True

        logger.info("[aps] decryption failure - no key found.")
        return pdu, False

    @source('nwk', 'NLDE-DATA')
    def on_nlde_data(
                        self,
                        nsdu,
                        destination_address_mode,
                        destination_address,
                        source_address,
                        security_use,
                        link_quality
    ):
        # Check that we got an App Data Payload here
        if ZigbeeAppDataPayload in nsdu:
            # Verify the security status
            security_status = (
                                APSSecurityStatus.SECURED_NWK_KEY if
                                security_use else
                                APSSecurityStatus.UNSECURED
            )
            # Check if a Security Header is present and attempt to decrypt packet if needed
            if (
                ZigbeeSecurityHeader in nsdu and
                nsdu[ZigbeeSecurityHeader].underlayer.__class__ is ZigbeeAppDataPayload
            ):
                security_status = APSSecurityStatus.SECURED_LINK_KEY
                decrypted, success = self.decrypt(nsdu)

                if success:
                    nsdu = decrypted
                else:
                    return

            # According to APS frametype, forward to the right service
            if nsdu.aps_frametype == 0: # data
                self.get_service("data").on_data_apdu(
                    nsdu,
                    destination_address_mode,
                    destination_address,
                    source_address,
                    security_status,
                    link_quality
                )
            elif nsdu.aps_frametype == 1: # command
                self.get_service("management").on_command_apdu(
                    nsdu,
                    destination_address_mode,
                    destination_address,
                    source_address,
                    security_status,
                    link_quality
                )
            elif nsdu.aps_frametype == 2: # ack
                pass

    @source('nwk', 'NLME-JOIN')
    def on_join(
                        self,
                        network_address,
                        extended_address,
                        capability_information,
                        rejoin=False,
                        secure_rejoin=False
    ):
        self.get_service("management").indicate_join(
            network_address,
            extended_address,
            capability_information,
            rejoin=rejoin,
            secure_rejoin=secure_rejoin
        )

    @source('nwk', 'INTRP-DATA')
    def on_intrp_data(
                        self,
                        asdu,
                        profile_id=0,
                        cluster_id=0,
                        destination_pan_id=0xFFFF,
                        destination_address=0xFFFF,
                        source_pan_id=0xFFFF,
                        source_address=0xFFFF,
                        link_quality=255
    ):
        """
        Callback processing NWK InterPAN Data indication.
        """
        self.get_service("interpan").indicate_interpan_data(
            asdu,
            profile_id=profile_id,
            cluster_id=cluster_id,
            destination_pan_id=destination_pan_id,
            destination_address=destination_address,
            source_pan_id=source_pan_id,
            source_address=source_address,
            link_quality=link_quality
        )

APSManager.add(APLManager)



"""
WHAD Enhanced ShockBurst injection configuration

This module defines the ESB configuration class for
packet injection.
"""
from dataclasses import dataclass

@dataclass
class InjectionConfiguration:
    """
    Configuration for injecting in an Enhanced ShockBurst communication.

    :param channel: select the channel to use (c)
    :param address: provide address to use (f)
    :param synchronize: enable synchronization (s)
    """
    channel : int = 0
    address : str = None
    synchronize : bool = False



"""WHAD Mock device for testing purpose.

This module provides a template class to design mock devices. Mock devices are
special devices intended to be used for unit testing, and should not be made
available to users.
"""
import logging
from threading import Thread, Event
from queue import Empty

from typing import Optional

from whad.hub import ProtocolHub
from whad.hub.message import HubMessage
from whad.hub.discovery import DeviceType

from ..device import Device, DeviceInfo, Disconnected, WhadDeviceDisconnected, \
    WhadDeviceNotReady

logger = logging.getLogger(__name__)

class MockInThread(Thread):
    """Internal thread processing data sent by the connector to the
    hardware interface.
    """

    def __init__(self, device: "MockDevice"):
        super().__init__()
        self.daemon = True
        self.__iface = device
        self.__canceled = False

    def cancel(self):
        """Cancel thread
        """
        self.__canceled = True

    def run(self):
        """Out thread main task.
        """
        while not self.__canceled:
            # Wait for a message to send to interface (blocking)
            logger.debug("[%s][in_thread] waiting for message to send", self.__iface.interface)
            try:
                with self.__iface.get_pending_message(timeout=1.0) as message:
                    # Read data from device (may block)
                    try:
                        logger.debug("[%s][in_thread] sending message %s",
                                     self.__iface.interface, message)

                        # Notify the device that we received a specific message from
                        # our connector:
                        self.__iface.on_connector_message(message)

                        # Notify message has correctly been sent, from a dedicated
                        # thread.
                        if message.has_callback():
                            Thread(target=message.sent).start()
                    except WhadDeviceNotReady:
                        if message.has_callback():
                            Thread(target=message.error, args=[1]).start()

                        # Exit processing loop
                        return
                    except WhadDeviceDisconnected:
                        if message.has_callback():
                            Thread(target=message.error, args=[2]).start()

                        # Exit processing loop
                        return
            except Empty:
                pass

class MockOutThread(Thread):
    """Internal thread processing data sent by the hardware interface
    to the device object.
    """

    def __init__(self, device: "MockDevice"):
        super().__init__()
        self.daemon = True
        self.__iface = device
        self.__canceled = False

    def cancel(self):
        """Cancel thread
        """
        self.__canceled = True

    def run(self):
        """Out thread main task.
        """
        while not self.__canceled:
            # Read data from device (may block)
            try:
                # Check if our mock device needs to fake a message sent by our
                # device.
                message = self.__iface.on_interface_message()
                if message is not None:
                    # Register message into the device queue
                    self.__iface.put_message(message)
            except WhadDeviceNotReady:
                break
            except WhadDeviceDisconnected:
                # Device has disconnected, notify interface by injecting a
                # `Disconnected` event message in its output message queue.
                # This event message will be filtered out by the Interface class
                # when called from our connector thread, and it will exit
                # gracefully
                logger.debug("[iface][%s] Device disconnected, sending event to connector.",
                             self.__iface.interface)
                self.__iface.put_message(Disconnected(self.__iface))
                break

class MockDevice(Device):
    """Mock device

    This device class is tied to no hardware and is intended to emulate the
    behavior of some real hardware device for a set of supported domains and
    features.
    """

    class route:
        """Mock device message routing decorator

        This decorator registered the decorated function as a callback for a
        specific message type.
        """

        def __init__(self, message_type):
            """Initialize decorator."""
            self.__msg_type = message_type

        def __call__(self, callback):
            """Called to decorate a specific callback"""
            setattr(callback, "_MESSAGE_TYPE", self.__msg_type)
            return callback


    # No interface to avoid instantiation
    INTERFACE_NAME = "mock"

    def __init__(self, author: str = 'whad', url: str = 'https://whad.io', proto_minver: int = 2,
                 version: str = '1.0', dev_type: int = DeviceType.VirtualDevice,
                 dev_id: bytes = b'', capabilities: Optional[dict] = None, max_speed: int = 115200,
                 index: Optional[int] = None):
        """Constructor."""
        # Loop over each method and registers those decorated with @route()
        self.__handlers = {}
        for prop_name in dir(self):
            try:
                prop_obj = getattr(self, prop_name)
                # If property is a method
                if callable(prop_obj) and hasattr(prop_obj, "_MESSAGE_TYPE"):
                    self.__handlers[getattr(prop_obj, "_MESSAGE_TYPE")] = prop_obj
            except AttributeError:
                pass

        # Set discovery properties
        self.__author = author
        self.__url = url
        self.__proto_minver = proto_minver
        self.__version = version
        self.__dev_type = dev_type
        self.__dev_id = dev_id
        self.__max_speed = max_speed
        self.__capabilities = capabilities or {}

        # Create a basic device
        super().__init__(index)

        # Custom device information management
        self.__info = None
        self.__iface_in = None
        self.__iface_out = None
        self.__hub = ProtocolHub()
        self.__discovered = False
        self.__opened = False

        # Waiting event
        self.__blocking_event = Event()

    @property
    def info(self) -> Optional[DeviceInfo]:
        """Device information object."""
        return self.__info

    @property
    def hub(self) -> ProtocolHub:
        """Retrieve the device protocol hub (parser/factory)
        """
        return self.__hub

    @property
    def opened(self) -> bool:
        """True if the mock device is open."""
        return self.__opened

    ##
    # Device operation
    ##

    def __start_io_threads(self):
        """Start background IO threads
        """
        self.__iface_in = MockInThread(self)
        self.__iface_in.start()
        self.__iface_out= MockOutThread(self)
        self.__iface_out.start()

    def __stop_io_threads(self):
        """Stop background IO threads
        """
        if self.__iface_in is not None:
            self.__iface_in.cancel()
            #self.__iface_in.join()
        if self.__iface_out is not None:
            self.__iface_out.cancel()
            #self.__iface_out.join()

    def open(self):
        """Open the mock device. We don't call Device.open() to avoid background
        threads to be created.
        """
        # Start background threads
        self.__start_io_threads()

        # Mark as opened
        self.__opened = True

        # Reset device
        self.reset()

    def reset(self) -> bool:
        """Reset device. We usually trigger a reset sequence and wait for a
        DeviceReady message.

        Override this method to implement a specific reset behavior.
        """
        return True

    def read(self) -> bytes:
        """Mock device does not emulate an hardware device by default, so read()
        is not supposed to be called.
        """
        logger.debug("[%s] unexpectedly reading bytes from interface.")
        return b""

    def write(self, payload: bytes) -> int:
        """Mock device does not emulate an hardware device by default, so write()
        is not supposed to be called.
        """
        logger.debug("[%s] unexpectedly writing bytes to interface: %s", self.interface,
                     payload.hex())
        return 0

    def close(self):
        """Close device.
        """
        # Stop io threads
        self.__stop_io_threads()

        # Mark device as closed
        self.__opened = False

        # Notify connector (if any) that the device has disconnected
        # Notify connector that device has closed
        if self.connector is not None:
            logger.debug("[%s] Send disconnection event to connector %s", self.interface,
                         self.connector)
            self.connector.send_event(Disconnected(self))
            logger.debug("[%s] Disconnection event sent !", self.interface)

    ##
    # Message hooks
    ##

    def on_connector_message(self, message: HubMessage):
        """Called whenever our output thread processes a message sent by a
        connector.

        :param message: Message to process
        :type message: HubMessage
        """
        # Do we have a haself.__handlers[type(message)]ndler registered for this specific message ?
        if type(message) in self.__handlers:
            logger.debug("[mock] Found handler associated with message type %s: %s", type(message), self.__handlers[type(message)])

            # Call message handler
            resp = self.__handlers[type(message)](message)

            # If handler returned a single message or a list of messages,
            # send them to the connector
            if isinstance(resp, HubMessage):
                self.put_message(resp)
            elif isinstance(resp, list):
                for m in resp:
                    if isinstance(m, HubMessage):
                        self.put_message(m)



    def on_interface_message(self) -> Optional[HubMessage]:
        """Called to check if the hardware interface needs to report something.
        This callback is mostly blocking by default, most of the processing
        using directly the device's `put_message` method to enqueue a message
        for the connector.

        However, it could be used to send notifications at a custom pace, if
        needed, depending on the subclass.
        """
        self.__blocking_event.wait()
        return None

    ##
    # Discovery management
    ##

    def discover(self):
        """Performs device discovery.

        Device discovered information is populated with the details provided
        when instantiating this device.
        """
        if not self.__discovered:
            # Convert capabilities dict into a domain/capabilities list
            caps = [ domain | cap[0] for domain, cap in self.__capabilities.items() ]

            # Create a new device information structure
            self.__info = DeviceInfo.create(
                author=self.__author,
                url=self.__url,
                version=self.__version,
                proto_ver=self.__proto_minver,
                max_speed=self.__max_speed,
                dev_type=self.__dev_type,
                dev_id=self.__dev_id,
                capabilities = caps
            )

            # Populate commands per domain
            for domain, cap in self.__capabilities.items():
                supp_commands = 0
                for cmd in cap[1]:
                    supp_commands = supp_commands | (1 << cmd)
                self.__info.add_supported_commands((domain&0xff000000), supp_commands)

            # Mark as discovered
            self.__discovered = True

        self.__hub = ProtocolHub(self.__proto_minver)

        # Set max transport speed
        if self.__info is not None:
            self.change_transport_speed(
                self.__info.max_speed
            )

    def get_domains(self) -> list[int]:
        """Get device' supported domains.

        :returns: list of supported domains
        :rtype: list
        """
        if self.__info is not None:
            return self.__info.domains

        # No domain discovered yet
        return []


    def get_domain_capability(self, domain):
        """Get a device domain capabilities.

        :param Domain domain: Target domain
        :returns: Domain capabilities
        :rtype: DeviceDomainInfoResp
        """
        if self.__info is not None:
            return self.__info.get_domain_capabilities(domain) or 0

        # No capability if not discovered
        return 0

    def get_domain_commands(self, domain):
        """Get a device supported domain commands.

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        :rtype: int
        """
        if self.__info is not None:
            return self.__info.get_domain_commands(domain)

        # No supported commands by default
        return 0

    def has_domain(self, domain) -> bool:
        """Query device for domain support."""
        if self.__info is not None:
            return self.__info.has_domain(domain)
        return False




"""
ReadByGroupType procedure

The ReadByGroupType procedure is initiated by a ReadByGroupType request, and
based on the required attribute start/end handle and group type returns a
ReadByGroupType response or an error.
"""
from typing import List
from struct import pack

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Read_By_Group_Type_Request, ATT_Read_By_Group_Type_Response

from whad.ble.profile.attribute import UUID
from whad.ble.stack.att.constants import BleAttErrorCode, BleAttOpcode
from whad.ble.stack.gatt.attrlist import GattAttributeDataList, GattGroupTypeItem

from .attribute import find_attr_by_type
from .procedure import BleClientProcedure, BleServerProcedure, UnexpectedProcError

class ServerReadByGroupTypeProcedure(BleServerProcedure):
    """ATT server ReadByGroupType procedure."""

    # Procedure operation code
    OPCODE = BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerReadByGroupType procedure."""
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure is triggered."""
        return ATT_Read_By_Group_Type_Request in request

    def process_request(self, request: Packet) -> list[Packet]:
        """React only on a ReadByGroupType request."""
        # We should not be called when request does not contain a ReadByGroupType request.
        if ATT_Read_By_Group_Type_Request not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        # Extract ReadByGroupType request
        request = request[ATT_Read_By_Group_Type_Request]

        try:
            # Make sure start handle is valid
            if request.start > request.end:
                self.set_state(self.states.DONE)
                return self.att_error_response(handle=request.start, ecode=BleAttErrorCode.INVALID_HANDLE)

            # List attributes corresponding to the provided group type, return an error
            # if no attribute is found.
            attrs = find_attr_by_type(self.attributes, UUID(request.uuid),
                                      start_handle=request.start, end_handle=request.end)
            if len(attrs) == 0:
                self.set_state(self.states.DONE)
                return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

            # Build response value
            resp = b""
            attr_data_size = None
            for attr in attrs:
                # Append attribute info to our response
                attr_data = pack("<HH", attr.handle, attr.end_handle) + attr.value

                # Save attribute data size if not set
                if attr_data_size is None:
                    attr_data_size = len(attr_data)

                # Append data if same size and not exceeding mtu
                if len(attr_data) == attr_data_size:
                    if self.mtu - len(resp) - 2 >= attr_data_size:
                        resp += attr_data
                    else:
                        break
                else:
                    break

            # Mark procedure as done
            self.set_state(self.states.DONE)

            # Return attribute list
            return [ ATT_Read_By_Group_Type_Response(
                    length=attr_data_size,
                    data=resp
                )
            ]

        except IndexError:
            self.set_state(self.states.DONE)
            return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

class ClientReadByGroupTypeProcedure(BleClientProcedure):
    """ATT client ReadByGroupType procedure."""

    # Procedure operation code
    OPCODE = BleAttOpcode.READ_BY_GROUP_TYPE_RESPONSE

    def __init__(self, group_type: UUID, start: int, end: int):
        """Initialize our ClientReadByGroupType procedure."""

        # Save requested parameters
        self.__uuid = group_type.value()
        self.__start = start
        self.__end = end

        # Initialize our procedure object, no attributes required
        # as we don't expose any attribute.
        super().__init__([
            ATT_Read_By_Group_Type_Request(
                uuid=self.__uuid,
                start=self.__start,
                end=self.__end
            )
        ])

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure is triggered."""
        return ATT_Read_By_Group_Type_Response in request

    def process_request(self, request: Packet) -> List[Packet]:
        """React only on a ReadByGroupType response."""
        # Do we received an ATT Error message ?
        if ATT_Error_Response in request:
            self.set_state(self.states.ERROR)
            self.set_result(request[ATT_Error_Response])
            return []

        # We should not be called when request does not contain a ReadByGroupType request.
        if ATT_Read_By_Group_Type_Response not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        # Extract ReadByGroupType response
        request = request[ATT_Read_By_Group_Type_Response]

        if request.length > 0:
            # Parse response
            items = GattAttributeDataList.from_bytes(
                request.data,
                request.length,
                GattGroupTypeItem
            )
            self.set_result(items)
            self.set_state(self.states.DONE)
        else:
            self.set_result(None)
            self.set_state(self.states.ERROR)

        # No packet to send.
        return []



from whad.zigbee.connector.enddevice import EndDevice

def create_enddevice(app, piped=False):
    enddevice = None
    # Is app stdin piped ?
    if piped:
        pass

    else:
        enddevice = EndDevice(app.interface)
    print("create enddevice")
    return (enddevice, None)



"""ZigBee Scan command handler
"""

from prompt_toolkit import print_formatted_text, HTML
from whad.cli.app import command
from whad.dot15d4.address import Dot15d4Address
from whad.zigbee.connector.enddevice import EndDevice

@command('scan')
def scan_handler(app, command_args):
    """scan for ZigBee networks

    <ansicyan><b>scan</b></ansicyan>

    This command will scan for Zigbee networks and show them in a list.
    """
    # We need to have an interface specified
    if app.is_piped_interface():
        app.error('This command cannot be used chained with another whad tool.')
    elif app.interface is not None:
        # Switch to Zigbee End Device mode
        enddevice = EndDevice(app.interface)

        print_formatted_text(HTML('<ansigreen>Channel   PAN ID   Ext. PAN ID             Joining</ansigreen>'))
        enddevice.start()
        try:
            for network in enddevice.discover_networks():
                # Show network
                print(
                        network.channel," "*6,
                        hex(network.pan_id)," ",
                        Dot15d4Address(network.extended_pan_id),
                        "permitted" if network.is_joining_permitted() else "forbidden"
                )

        except KeyboardInterrupt as keybd_int:
            print('\rScan terminated by user')
        enddevice.stop()
    else:
        app.error('You need to specify an interface with option --interface.')



"""ZigBee Applicative Layer (APL) exceptions
"""


class APLTimeoutException(Exception):
    """
    This exception is triggered when a timeout occurred at the APL layer.
    """
    def __init__(self):
        super().__init__()

class APLInvalidAddress(Exception):
    """
    This exception is triggered when an invalid address is detected.
    """
    def __init__(self):
        super().__init__()



"""RF4CE Target connector
"""
import logging

from whad.exceptions import UnsupportedCapability
from whad.dot15d4.stack import Dot15d4Stack
from whad.dot15d4.stack.mac import MACManager
from whad.dot15d4.address import Dot15d4Address

from .base import RF4CE
from ..stack.nwk import NWKManager

logger = logging.getLogger(__name__)

class Target(RF4CE):
    """
    RF4CE Target Node interface for compatible WHAD device.
    """
    def __init__(self, device, profiles=[]):
        RF4CE.__init__(self, device)

        # Check if device can act as target node

        if not self.can_be_coordinator():
            raise UnsupportedCapability("Target")

        # Stack initialization
        MACManager.add(NWKManager)
        self.__stack = Dot15d4Stack(self)

        # Channel initialization
        self.__channel = 15
        self.__channel_page = 0

        self.__stack.get_layer('mac').set_extended_address(Dot15d4Address("11:22:33:44:55:66:77:88").value)
        self.__stack.get_layer('mac').set_short_address(Dot15d4Address(0x0384).value)
        self.__stack.get_layer('mac').database.set("macPanId", 0x1234)
        #self.__stack.get_layer('mac').database.set("macAckTimeout", 0.5)

        for profile in profiles:
            self.__stack.get_layer('apl').add_profile(profile)

        self.enable_reception()


    @property
    def stack(self):
        return self.__stack

    def enable_reception(self):
        self.set_coordinator_mode(channel=self.__channel)

    def set_channel(self, channel=11):
        self.__channel = channel
        self.enable_reception()

    def perform_ed_scan(self, channel):
        if not self.can_perform_ed_scan():
            raise UnsupportedCapability("EnergyDetection")
        self.__channel = channel
        super().perform_ed_scan(channel)

    def set_channel_page(self, page=0):
        if page != 0:
            raise UnsupportedCapability("ChannelPageSelection")
        else:
            self.__channel_page = page

    def auto_discovery(self):
        self.__stack.get_layer('nwk').get_service('management').auto_discovery()

    def discovery_response(self, status, destination_address):
        self.__stack.get_layer('nwk').get_service('management').discovery_response(
            status,
            Dot15d4Address(destination_address).value
        )



    def get_channel(self):
        return self.__channel

    def get_channel_page(self):
        return self.__channel_page


    def send(self, packet):
        packet.reserved = 1
        packet.do_build()
        super().send(packet, channel=self.__channel)

    def on_pdu(self, pdu):
        if (
            hasattr(pdu,"metadata") and
            hasattr(pdu.metadata, "is_fcs_valid") and
            not pdu.metadata.is_fcs_valid
        ):
            return

        self.__stack.on_pdu(pdu)

    def on_ed_sample(self, timestamp, sample):
        self.__stack.on_ed_sample(timestamp, sample)



"""Bluetooth Low Energy traffic analyzers.

This module provides multiple traffic analyzers required by `wanalyze` to process
and analyze BLE packets from PCAP files or live capture. These analyzers are basic
state machines that are fed with packets and return inferred/recovered information.
"""
import json
from struct import unpack

from scapy.layers.bluetooth4LE import BTLE_ADV_IND, BTLE_SCAN_RSP, BTLE_CONNECT_REQ
from scapy.layers.bluetooth import  ATT_Read_By_Group_Type_Request, \
    ATT_Read_By_Group_Type_Response, ATT_Read_By_Type_Request, \
    ATT_Read_By_Type_Response, ATT_Error_Response, ATT_Find_Information_Request, \
    ATT_Find_Information_Response

from whad.common.analyzer import TrafficAnalyzer
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.stack.gatt.message import GattReadByGroupTypeResponse, GattReadByTypeResponse, \
    GattFindInfoResponse
from whad.ble.profile.attribute import UUID
from whad.ble.profile import GenericProfile
from whad.ble.profile.characteristic import Characteristic, CharacteristicDescriptor, \
    ClientCharacteristicConfig
from whad.ble.profile.service import PrimaryService, SecondaryService
from whad.ble.crypto import EncryptedSessionInitialization, LegacyPairingCracking, \
    LongTermKeyDistribution, IdentityResolvingKeyDistribution, \
    ConnectionSignatureResolvingKeyDistribution

class PeripheralInformation(TrafficAnalyzer):
    """Peripheral information

    This class processes packets and extract information about devices.
    """

    def __init__(self):
        self.adv_ind_list = []
        self.scan_rsp_list = []
        self.adv_data = None
        self.scan_rsp = None
        self.address = None
        super().__init__()

    def reset(self):
        """Reset this traffic analyzer.
        """
        super().reset()
        self.adv_ind_list = []
        self.scan_rsp_list = []
        self.adv_data = None
        self.scan_rsp = None
        self.address = None

    @property
    def output(self) -> dict:
        """Analyzer output
        """
        if self.scan_rsp is None:
            return {
                "adv_data" : self.adv_data,
                "bd_addr" : str(self.address),
                "addr_type" : self.address.type if self.address is not None else None
            }

        # Include scan response
        return {
            "adv_data" : self.adv_data,
            "scan_rsp" : self.scan_rsp,
            "bd_addr" : str(self.address),
            "addr_type" : self.address.type if self.address is not None else None
        }


    def process_packet(self, packet):
        """Process a specific packet and modify the state of this analyzer.
        """
        if BTLE_ADV_IND in packet:
            self.trigger()
            self.adv_ind_list.append(packet)
        elif BTLE_SCAN_RSP in packet:
            self.trigger()
            self.scan_rsp_list.append(packet)
        elif BTLE_CONNECT_REQ in packet:
            self.trigger()
            self.address = BDAddress(packet.AdvA, random=packet.TxAdd != 0)
            for adv_ind in self.adv_ind_list[::-1]:
                if adv_ind.AdvA == packet.AdvA:
                    self.adv_data = b"".join([bytes(i) for i in adv_ind.data])
                    break
            for scan_rsp in self.adv_ind_list[::-1]:
                if scan_rsp.AdvA == packet.AdvA:
                    self.scan_rsp = b"".join([bytes(i) for i in scan_rsp.data])
                    break
            if self.scan_rsp is not None and self.adv_data is not None:
                self.complete()

class ReadByGroupTypeDiscovery(TrafficAnalyzer):
    """GATT services and characteristics discovery traffic analyzer.
    """

    def __init__(self):
        self.uuid = None
        self.handle = None
        self.start_handle = None
        self.end_handle = None
        self.items = []
        super().__init__()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.uuid = None
        self.handle = None
        self.start_handle = None
        self.end_handle = None
        self.items = []

    @property
    def output(self):
        """Current analyzer output state.
        """
        return {
            "uuid" : self.uuid,
            "handle" : self.handle,
            "start_handle" : self.start_handle,
            "end_handle" : self.end_handle,
            "items" : self.items
        }

    def process_packet(self, packet):
        """Modify analyzer state by processing a specific packet.
        """
        if ATT_Read_By_Group_Type_Request in packet:

            if not self.triggered:
                self.start_handle = packet.start
                self.end_handle = packet.end
                self.uuid = packet.uuid
                self.trigger()
            else:
                if packet.end != self.end_handle or packet.uuid != self.uuid:
                    if len(self.items) > 0:
                        self.complete()

                    self.reset()
            self.mark_packet(packet)

        elif ATT_Read_By_Group_Type_Response in packet and self.triggered:
            response = GattReadByGroupTypeResponse.from_bytes(
                item_size = packet.length,
                data = packet.data
            )
            for item in response:
                self.items.append(item)
                self.handle = item.end + 1
                if item.end in  (0xFFFF, self.end_handle):
                    self.complete()

            self.mark_packet(packet)

        elif ATT_Error_Response in packet and packet.request == 0x10 and \
            packet.ecode == 0x0a and self.triggered:
            self.mark_packet(packet)
            self.complete()



class ReadByTypeDiscovery(TrafficAnalyzer):
    """GATT read by type traffic analyzer.
    """

    def __init__(self):
        self.uuid = None
        self.handle = None
        self.start_handle = None
        self.end_handle = None
        self.items = []
        super().__init__()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.uuid = None
        self.handle = None
        self.start_handle = None
        self.end_handle = None
        self.items = []


    @property
    def output(self):
        """Output state.
        """
        return {
            "uuid" : self.uuid,
            "handle" : self.handle,
            "start_handle" : self.start_handle,
            "end_handle" : self.end_handle,
            "items" : self.items
        }

    def process_packet(self, packet):
        """Modify analyzer state by processing a specific packet.
        """
        if ATT_Read_By_Type_Request in packet:
            if not self.triggered:
                self.start_handle = packet.start
                self.end_handle = packet.end
                self.uuid = packet.uuid
                self.trigger()
            else:
                if packet.end != self.end_handle or packet.uuid != self.uuid:
                    self.reset()
                    self.start_handle = packet.start
                    self.end_handle = packet.end
                    self.uuid = packet.uuid
                    self.trigger()
            self.mark_packet(packet)

        elif ATT_Read_By_Type_Response in packet and self.triggered:
            response = GattReadByTypeResponse.from_bytes(
                item_size = packet[ATT_Read_By_Type_Response].len,
                data = bytes(packet[ATT_Read_By_Type_Response:])[1:]
            )
            for item in response:
                self.items.append(item)
            self.complete()
            self.mark_packet(packet)
        elif ATT_Error_Response in packet and packet.request == 0x08 and \
            packet.ecode == 0x0a and self.triggered:
            self.complete()
            self.mark_packet(packet)
        else:
            pass

class FindInformationDiscovery(TrafficAnalyzer):
    """GATT device information discovery analyzer.
    """

    def __init__(self):
        self.start_handle = None
        self.end_handle = None
        self.items = []
        super().__init__()

    def reset(self):
        """Reset analyzer.
        """
        super().reset()
        self.start_handle = None
        self.end_handle = None
        self.items = []


    @property
    def output(self):
        """Current analyzer state.
        """
        return {
            "start_handle" : self.start_handle,
            "end_handle" : self.end_handle,
            "items" : self.items
        }

    def process_packet(self, packet):
        """Modify traffic analyzer state by processing a specific packet.
        """
        if ATT_Find_Information_Request in packet:
            if not self.triggered:
                self.start_handle = packet.start
                self.end_handle = packet.end
                self.trigger()
            else:
                if packet.end != self.end_handle:
                    self.reset()
                    self.start_handle = packet.start
                    self.end_handle = packet.end
                    self.trigger()
            self.mark_packet(packet)
        elif ATT_Find_Information_Response in packet and self.triggered:
            response = GattFindInfoResponse.from_bytes(
                format = packet[ATT_Find_Information_Response].format,
                data = bytes(packet[ATT_Find_Information_Response:])[1:]
            )
            for item in response:
                self.items.append(item)
            self.complete()
            self.mark_packet(packet)
        elif ATT_Error_Response in packet and packet.request == 0x04 and self.triggered:
            self.complete()
            self.mark_packet(packet)

class PrimaryServicesDiscovery(ReadByGroupTypeDiscovery):
    """GATT primary services discovery analyzer.
    """

    def __init__(self):
        self.primary_services = []
        super().__init__()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.primary_services = []


    @property
    def output(self):
        """Current analyzer state.
        """
        return {
            "primary_services" : self.primary_services,
        }

    def complete(self):
        """Update output state.
        """
        if self.uuid == 0x2800:
            self.primary_services = [
                PrimaryService(
                    uuid=UUID(item.value),
                    handle=item.handle,
                    end_handle=item.end
                )
                for item in self.items
            ]
            super().complete()
        else:
            self.reset()


class CharacteristicsDiscovery(ReadByTypeDiscovery):
    """GATT characterstics discovery analyzer.
    """

    def __init__(self):
        self.characteristics = []
        super().__init__()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.characteristics = []

    @property
    def output(self):
        """Current analyzer output.
        """
        return {
            "characteristics" : self.characteristics,
        }

    def complete(self):
        """Update output state.
        """
        if self.uuid == 0x2803:
            for item in self.items:
                charac_properties = item.value[0]
                charac_handle = item.handle
                charac_value_handle = unpack('<H', item.value[1:3])[0]
                charac_uuid = UUID(item.value[3:])
                charac = Characteristic(
                    uuid=charac_uuid,
                    properties=charac_properties
                )
                charac.handle = charac_handle
                charac.value_handle = charac_value_handle
                self.characteristics.append(charac)
            super().complete()
        else:
            self.reset()

class SecondaryServicesDiscovery(ReadByGroupTypeDiscovery):
    """GATT secondary services discovery analyzer.
    """

    def __init__(self):
        self.secondary_services = []
        super().__init__()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.secondary_services = []

    @property
    def output(self):
        """Current analyzer output state.
        """
        return {
            "secondary_services" : self.secondary_services,
        }

    def complete(self):
        """Update output state.
        """
        if self.uuid == 0x2801:
            self.secondary_services = [
                SecondaryService(
                    uuid=UUID(item.value),
                    handle=item.handle,
                )
                for item in self.items
            ]
            super().complete()
        else:
            self.reset()

class ServicesDiscovery(TrafficAnalyzer):
    """GATT services discovery.

    This is a high-level analyzer that uses :class:`PrimaryServicesDiscovery` and
    :class:`SecondaryServicesDiscovery` analyzers.
    """

    def __init__(self):
        self.primary_service_analyzer = None
        self.secondary_service_analyzer = None
        self.services = []
        super().__init__()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.primary_service_analyzer = PrimaryServicesDiscovery()
        self.secondary_service_analyzer = SecondaryServicesDiscovery()
        self.services = []

    @property
    def output(self):
        """Output state.
        """
        return {
            "services" : self.services,
        }

    def process_packet(self, packet):
        """Update sub-analyzers by processing a given packet.
        """
        self.primary_service_analyzer.process_packet(packet)
        self.secondary_service_analyzer.process_packet(packet)

        if self.primary_service_analyzer.triggered:
            self.trigger()

        if self.primary_service_analyzer.completed:
            for pkt in self.primary_service_analyzer.marked_packets:
                self.mark_packet(pkt)
            self.services += self.primary_service_analyzer.primary_services
            self.complete()

        if self.secondary_service_analyzer.completed:
            for pkt in self.secondary_service_analyzer.marked_packets:
                self.mark_packet(pkt)
            self.services += self.secondary_service_analyzer.secondary_services
            self.complete()

class DescriptorsDiscovery(FindInformationDiscovery):
    """GATT descriptors discovery analyzer.
    """

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.descriptors = []


    @property
    def output(self):
        """Output state
        """
        return {
            "descriptors" : self.descriptors,
        }

    def complete(self):
        """Update output state.
        """
        for item in self.items:
            if item.uuid  == UUID(0x2901) or item.uuid == UUID(0x2902):
                self.descriptors.append(item)

        if len(self.descriptors) > 0:
            super().complete()


class GATTServerDiscovery(TrafficAnalyzer):
    """GATT Server discovery analyzer.

    This analyzer relies on GATT primary/secondary service analyzers as well as
    characteristic and descriptor analyzer to rebuild dynamically a device GATT
    profile from captured packets.
    """

    class InferredProfile(GenericProfile):
        """GATT profile inferred from traffic analysis.
        """

        def __init__(self):
            super().__init__()
            self.devinfo = None

        def export_json(self) -> str:
            """Export inferred profile to JSON file.
            """
            json_value = super().export_json()
            if hasattr(self, "devinfo"):
                profile = json.loads(json_value)
                profile["devinfo"] = self.devinfo
                for k in profile["devinfo"].keys():
                    if isinstance(profile["devinfo"][k], bytes):
                        profile["devinfo"][k] = profile["devinfo"][k].hex()
                json_value = json.dumps(profile)
            return json_value

    def __init__(self):
        self.services_completed = False
        self.characteristics_completed = False
        self.profile = None
        super().__init__()

    def reset(self):
        """Reset analyzer.
        """
        super().reset()
        self.primary_service_analyzer = PrimaryServicesDiscovery()
        self.secondary_service_analyzer = SecondaryServicesDiscovery()
        self.characteristics_analyzer = CharacteristicsDiscovery()
        self.descriptors_discovery = DescriptorsDiscovery()
        self.peripheral_information = PeripheralInformation()

        self.services_completed = False
        self.characteristics_completed = False

        self.profile = None
        self.services = []
        self.characteristics = []
        self.descriptors = []

    @property
    def output(self):
        """Output state
        """
        return {
            "profile" : self.profile
        }

    def complete(self):
        """Update output state
        """
        super().complete()
        self.profile = self.InferredProfile()
        for service in self.services:
            self.profile.add_service(service)
        if self.peripheral_information.completed:
            self.profile.devinfo = self.peripheral_information.output

    def process_packet(self, packet):
        """Update analyzers state by processing a packet.
        """
        self.primary_service_analyzer.process_packet(packet)
        self.secondary_service_analyzer.process_packet(packet)
        self.characteristics_analyzer.process_packet(packet)
        self.descriptors_discovery.process_packet(packet)
        self.peripheral_information.process_packet(packet)


        if self.primary_service_analyzer.triggered:
            self.trigger()

        if self.primary_service_analyzer.completed:

            for pkt in self.primary_service_analyzer.marked_packets:
                self.mark_packet(pkt)

            self.services += self.primary_service_analyzer.primary_services
            self.services_completed = True
            self.primary_service_analyzer.reset()

        if self.secondary_service_analyzer.completed:

            for pkt in self.secondary_service_analyzer.marked_packets:
                self.mark_packet(pkt)

            self.services += self.secondary_service_analyzer.secondary_services
            self.secondary_service_analyzer.reset()

        if self.characteristics_analyzer.completed:

            for pkt in self.characteristics_analyzer.marked_packets:
                self.mark_packet(pkt)

            for characteristic in self.characteristics_analyzer.characteristics:
                for service in self.services:
                    if service.handle < characteristic.handle and \
                            characteristic.handle <= service.end_handle:
                        service.add_characteristic(characteristic)
            self.characteristics_completed = True
            self.characteristics_analyzer.reset()

        if self.descriptors_discovery.completed:

            for pkt in self.descriptors_discovery.marked_packets:
                self.mark_packet(pkt)

            for descriptor in self.descriptors_discovery.descriptors:
                self.descriptors += self.descriptors_discovery.descriptors
                selected_service = None
                for service in self.services:
                    if service.handle < descriptor.handle and \
                            descriptor.handle <= service.end_handle:
                        selected_service = service
                        break

                if selected_service is not None:
                    characteristics = list(selected_service.characteristics())
                else:
                    characteristics = []
                selected_characteristic = None
                for characteristic in characteristics:
                    if characteristic.handle <= descriptor.handle:
                        selected_characteristic = characteristic
                    else:
                        break
                if selected_characteristic is not None:
                    descriptor_obj = None
                    if descriptor.uuid  == UUID(0x2901):
                        descriptor_obj = CharacteristicDescriptor(
                                descriptor.uuid,
                                handle=descriptor.handle
                        )
                    elif descriptor.uuid == UUID(0x2902):
                        descriptor_obj = ClientCharacteristicConfig(
                                descriptor.handle
                        )
                    if descriptor_obj is not None:
                        selected_characteristic.add_descriptor(descriptor_obj)

            self.descriptors_discovery.reset()


        if (self.triggered and
            self.characteristics_completed and
            not self.characteristics_analyzer.triggered and
            self.services_completed and
            not self.primary_service_analyzer.triggered and
            not self.descriptors_discovery.triggered
        ):
            handles = []
            for service in self.services:
                handles.append(service.handle)
                for characteristic in service.characteristics():
                    handles.append(characteristic.handle)
                    handles.append(characteristic.value_handle)

                    for descriptor in characteristic.descriptors():
                        handles.append(descriptor.handle)

                    if characteristic.can_notify() or characteristic.can_indicate():
                        if len(list(characteristic.descriptors())) == 0:
                            return

            if (
                    all([len(list(service.characteristics())) > 0 for service in self.services]) and
                    list(range(min(*handles), max(*handles)+1)) == handles
                ):
                self.complete()


analyzers = {
    "peripheral_information" : PeripheralInformation,
    "encrypted_session_initialization" : EncryptedSessionInitialization,
    "legacy_pairing_cracking" : LegacyPairingCracking,
    "ltk_distribution" : LongTermKeyDistribution,
    "irk_distribution" : IdentityResolvingKeyDistribution,
    "csrk_distribution" : ConnectionSignatureResolvingKeyDistribution,
    "profile_discovery" : GATTServerDiscovery
}



"""BLE profile command handler
"""
from json import loads, dumps
from argparse import Namespace

from prompt_toolkit import print_formatted_text, HTML

from whad.exceptions import WhadDeviceDisconnected
from whad.ble.profile.characteristic import CharacteristicProperties
from whad.cli.app import command, CommandLineApp
from whad.ble import Central, Scanner
from whad.ble.exceptions import ConnectionLostException, PeripheralNotFound
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.stack.att.exceptions import AttError
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.cli.central.helpers import show_att_error

# Expected parameters that must be passed to our program
# to use an already established connection
EXPECTED_BLE_PARAMS = [
    "initiator_bdaddr",
    "initiator_addrtype",
    "target_bdaddr",
    "target_addrtype",
    "conn_handle",
]

def profile_discover(app: CommandLineApp, device, include_values: bool = False) -> bool:
    """Discover the GATT profile of a given device

    :param app: WHAD application instance
    :type app: :class:ẁhad.cli.app.CommandLineApp`
    :param device: Peripheral device to enumerate
    :type device: :class:`whad.ble.profile.PeripheralDevice`
    :return: `True` on success, `False` if an error occurred during profile discovery
    :rtype: bool
    """
    try:
        device.discover(include_values=include_values)
    except AttError as atterr:
        show_att_error(app, atterr)
        return False
    except GattTimeoutException:
        app.error("GATT Timeout while discovering services and characteristics. Aborted.")
        return False

    # Show discovered services and characteristics
    for service in device.services():
        print_formatted_text(HTML((
            "<ansicyan><b>Service {name}</b></ansicyan> "
            " (handle {handle} to {end_handle})"
            )).format(
                name=service.name, handle=service.handle,
                end_handle=service.end_handle
            )
        )

        # Loop on service characteristics
        for charac in service.characteristics():
            properties = charac.properties
            rights = ''
            if properties & CharacteristicProperties.READ != 0:
                rights += '<ansigreen>R</ansigreen>'
            if properties & CharacteristicProperties.WRITE != 0:
                rights += '<ansired>W</ansired>'
            if properties & CharacteristicProperties.WRITE_WITHOUT_RESPONSE != 0:
                rights += '<ansimagenta>W</ansimagenta>'
            if properties & CharacteristicProperties.INDICATE != 0:
                rights += '<ansicyan>I</ansicyan>'
            if properties & CharacteristicProperties.NOTIFY != 0:
                rights += '<ansiblue>N</ansiblue>'

            # Print characteristic properties
            print_formatted_text(
                HTML("  <b>{name}</b> %s, handle {handle}, value handle: {value_handle}" % rights)
                    .format(
                        name=charac.name,
                        rights=rights,
                        handle=charac.handle,
                        value_handle=charac.value_handle
                    )
            )

            # Display descriptors
            for desc in charac.descriptors():
                print_formatted_text(
                    HTML("    <ansiblue>Descriptor type {name}</ansiblue>, handle: {handle}")
                        .format(name=desc.name, handle=desc.handle)
                )


            # Show values if requested
            if include_values and charac.cached_value is not None:
                print_formatted_text(
                    HTML("    <ansiblue>Value</ansiblue>: {value}")
                        .format(value=charac.cached_value.hex())
                )
        print('')

    # Success
    return True

def profile_export(app, command_args: list, device, device_metadata: dict):
    """Export a discovered GATT profile into a JSON file.
    """
    # Load GATT profile JSON data
    json_data = device.export_json()
    profile = loads(json_data)

    # Add specific device info (for emulating)
    profile['devinfo'] = device_metadata
    json_data = dumps(profile)
    try:
        print(f"Writing profile JSON data to {command_args[0]} ...")
        with open(command_args[0], 'w', encoding='utf-8') as profile:
            profile.write(json_data)
    except IOError:
        app.error("An error occurred when writing to %s", command_args[0])

@command('profile')
def profile_handler(app, command_args):
    """discover services and characteristics

    <ansicyan><b>profile</b> <i>[JSON_PROFILE]</i></ansicyan>

    This command connects to a target device, discover its services and characteristics,
    and optionnaly exports the device's profile into a <i>JSON_PROFILE</i> file. This JSON
    file may be used later for emulating a similar device with the exact same properties
    and profile, see <ansicyan>emulate</ansicyan> command.
    """
    # We need to have an interface specified
    if app.interface is not None and app.args.bdaddr is not None:

        print(f"Searching for target device {app.args.bdaddr} ...")

        # Switch to Scanner mode, search our device
        scanner = Scanner(app.interface)
        scanner.start()

        device = None
        for scanned_device in scanner.discover_devices(timeout=30.0, updates=True):
            if scanned_device.address.lower() == app.args.bdaddr.lower():
                device = scanned_device
                if device.got_scan_rsp:
                    break

        # Check device is our device
        if device is None:
            # Stop scanner
            scanner.stop()

            # Display error message
            app.error(f"BLE peripheral {app.args.bdaddr} cannot be found.")
            return

        # Generate device advertising information
        device_metadata = {
            "adv_data": device.adv_records.to_bytes().hex(),
            "bd_addr": str(device.address),
            "addr_type": device.address_type,
            "scan_rsp": None
        }
        if device.scan_rsp_records is not None:
            device_metadata["scan_rsp"] = device.scan_rsp_records.to_bytes().hex()

        scanner.stop()

        retries = 3
        json = None
        while retries > 0:
            # Switch to central mode
            central = Central(app.interface, from_json=json)
            central.start()

            try:
                print("Connecting to target device ...")

                # Connect to target device
                device = central.connect(app.args.bdaddr, random=app.args.random)
                if device is None:
                    app.error("Cannot connect to %s, device does not respond.", app.args.bdaddr)
                else:
                    try:
                        print("Enumerating services and characteristics ...")
                        # Perform profile discovery
                        result = profile_discover(app, device, include_values=(len(command_args) >= 1))

                        # Export profile to JSON file if required
                        if result and len(command_args) >= 1:
                            # Load GATT profile JSON data
                            json_data = device.export_json()
                            profile = loads(json_data)

                            # Add specific device info (for emulating)
                            profile['devinfo'] = device_metadata
                            json_data = dumps(profile)
                            try:
                                print(f"Writing profile JSON data to {command_args[0]} ...")
                                with open(command_args[0], 'w', encoding='utf-8') as profile:
                                    profile.write(json_data)
                            except IOError:
                                app.error(f"An error occured when writing to {command_args[0]}")

                        # Disconnect
                        device.disconnect()

                        # Exit while loop
                        retries = 0
                        break
                    except ConnectionLostException:
                        app.error("BLE device disconnected during discovery.")
                        if retries > 0:
                            # Try again and feed all the discovered information into the profile used in Central,
                            # to speed up discovery.
                            retries -= 1
                            json = device.export_json()
                            app.warning(f"BLE peripheral {app.args.bdaddr} has disconnected, retrying ({retries} attempt(s) left)")
                        else:
                            app.error(f"BLE peripheral {app.args.bdaddr} has disconnected, exiting.")
                            break

            except PeripheralNotFound:
                app.error(f"BLE peripheral {app.args.bdaddr} cannot be found.")
            except WhadDeviceDisconnected:
                if retries > 0:
                    retries -= 1
                    app.warning(f"BLE peripheral {app.args.bdaddr} has disconnected, retrying ({retries} attempt(s) left)")
                else:
                    app.error(f"BLE peripheral {app.args.bdaddr} has disconnected, exiting.")

            # Terminate central
            central.stop()

    # Piped interface
    elif app.args.bdaddr is None and app.is_piped_interface():

        # Make sure we have all the required parameters
        for param in EXPECTED_BLE_PARAMS:
            if not hasattr(app.args, param):
                app.error("Source interface does not provide a BLE connection")

        initiator = BDAddress(str(app.args.initiator_bdaddr),
                              addr_type=int(app.args.initiator_addrtype))
        advertiser = BDAddress(str(app.args.initiator_bdaddr),
                               addr_type=int(app.args.initiator_addrtype))
        existing_connection = Namespace(
            initiator=initiator.value,
            init_addr_type=int(app.args.initiator_addrtype),
            advertiser=advertiser.value,
            adv_addr_type=int(app.args.target_addrtype),
            conn_handle=int(app.args.conn_handle)
        )

        central = Central(app.input_interface, existing_connection)
        device = central.peripheral()

        try:
            # Read GATT characteristic
            profile_discover(app, device, include_values=False)

            # Export profile to JSON file if required
            if len(command_args) >= 1:
                app.warning("Cannot create profile file when chained.")

        except ConnectionLostException:
            app.error("BLE device disconnected during discovery.")
    else:
        app.error("You need to specify a target device with option --bdaddr.")



from whad.dot15d4.stack.database import Dot15d4Database
from whad.zigbee.stack.nwk.neighbors import NWKNeighborTable

class NWKIB(Dot15d4Database):
    """
    ZigBee NWKIB Database of attributes.
    """

    def reset(self):
        """
        Reset the NWKIB database to its default value.
        """
        self.nwkSequenceNumber = 0
        #self.nwkPassiveAckTimeout = None
        self.nwkMaxBroadcastRetries = 3
        self.nwkMaxChildren = 4
        self.nwkMaxDepth = 4 # ?
        self.nwkMaxRouters = 6
        self.nwkNeighborTable = NWKNeighborTable()
        #self.nwkNetworkBroadcastDeliveryTime = None
        self.nwkReportConstantCost = 0
        self.nwkRouteTable = []
        self.nwkSymLink = False
        self.nwkCapabilityInformation = 0
        self.nwkAddrAlloc = 0
        self.nwkUseTreeRouting = True
        self.nwkManagerAddr = 0
        self.nwkMaxSourceRoute = 0xc
        self.nwkUpdateId = 0
        self.nwkNetworkAddress = 0xFFFF
        self.nwkExtendedPANID = 0x0000000000000000
        self.nwkPANId = 0xFFFF
        self.nwkIeeeAddress = 0x6055f90000f714e4
        self.nwkLeaveRequestAllowed = True
        self.nwkTxTotal = 0

        self.nwkStackProfile = 2
        self.nwkcProtocolVersion = 2

        self.nwkSecurityLevel = 0
        self.nwkSecurityMaterialSet = []
        self.nwkActiveKeySeqNumber = 0
        self.nwkSecureAllFrames = False
        self.nwkAllFresh = False
        self.nwkLinkStatusPeriod = 0x0f
        self.nwkRouterAgeLimit = 3

        self.nwkParentInformation = 0
        self.nwkCapabilityInformation = None
        self.nwkAddressMap = {}

        self.nwkUseMulticast = True
        self.nwkOwnNetwork = None



"""WHAD Protocol Generic Progress message abstraction layer.
"""
from whad.hub.message import pb_bind, PbFieldInt, PbMessageWrapper
from whad.hub.generic import Generic

@pb_bind(Generic, 'progress', 1)
class Progress(PbMessageWrapper):
    """Generic progress message.
    """
    value = PbFieldInt('generic.progress.value')



"""
Devices
-------

WHAD provides various classes to interact with WHAD-enabled hardware:

- :py:class:`whad.device.device.Device`
- :py:class:`whad.device.device.VirtualDevice`

Class :class:`whad.device.device.Device` is the default class that allows WHAD devices
enumeration and access. It is the main class to use to open any device, through
its :py:meth:`whad.device.Device.create` method as shown below:

.. code-block:: python

    from whad.device import Device

    dev = Device.create("uart0")

The :py:class:`whad.device.device.VirtualDevice` shall not be directly used. This class
is used to add support for incompatible WHAD devices like the *Ubertooth*
or the *ApiMote* and acts as an adaptation layer between the underlying WHAD
protocol and the specific protocol used by the target hardware.

.. important::

    The :py:class:`whad.device.device.WhadDevice` class that is still defined in WHAD (and used
    in some old example scripts or documentation) is an alias for the new
    :py:class:`whad.device.device.Device` class, and is meant to be deprecated in the future.
    This old class has been renamed to ``Device`` for clarity, and the same happened
    with the old default connector class :py:class:`whad.device.connector.WhadConnector`
    that has been renamed to :py:class:`whad.device.connector.Connector`.

    These old classes will be marked as *deprecated* in a future release, with a
    specific EOL date announced. A warning message will be issued in case one of
    these classes is used in a script or a tool to give time to users to migrate
    to the new ones (renaming classes is enough to switch to the new implementation,
    APIs stay the same).


.. autoclass:: whad.device.device.Device
    :members:

.. autoclass:: whad.device.device.VirtualDevice
    :members:


Connectors
----------

*Connectors* shall ensure the device they are linked to does support the
target domain and a mimimal set of commands, and can tailor its behavior
depending on the capabilities of the hardware. If a *connector* is linked
to a device that either does not support the *domain* this *connector* is
supposed to operate or lacks specific *commands*, a
:py::class:`whad.exceptions.UnsupportedDomain` exception or a
:py:class:`whad.exceptions.UnsupportedCapability` may be raised.

WHAD provides a default connector class, :py:class:`whad.device.connector.Connector`,
that implements a set of features out-of-the-box:

- Packet and message sniffing and processing
- Event notification mechanism
- Synchronous mode

Sniffing packet and messages could be useful to implement packet sniffers or
intercept some specific events like disconnection of the linked hardware device.
Most of the time this feature is used to sniff packets related to a target domain.
The :py:meth:`whad.device.connector.Connector.sniff` method is specifically
tailored for this use. When not sniffing, packets received from the hardware device
are forwarded to the connector's packet processing methods than can be overriden by
inheriting classes.

By default, the default connector class provides methods to add and remove custom
event listeners (:py:meth:`whad.device.connector.Connector.add_listener` and
:py:meth:`whad.device.connector.Connector.remove_listener`), and an additional
method to send an event to the registered listeners (:py:meth:`whad.device.connector.Connector.notify`).

Last but not least, the provided *synchronous mode* will disable packet forwarding
and save all received packets in a reception queue, waiting for the application to
retrieve and process them. Service messages will still be processed by the *connector*,
in order to handle any device disconnection or other unexpected event that may occur.
When this *synchronous mode* is disabled, every unprocessed packet stored in the
reception queue are automatically forwarded to the connector's packet processing
methods, and will be then dispatched to the corresponding handlers.

.. autoclass:: whad.device.connector.Connector
    :members:

    .. automethod:: __init__

.. autoclass:: whad.device.connector.LockedConnector
    :members:

    .. automethod:: __init__

Deprecated classes
------------------

In early versions of WHAD, device and connector base classes had different names
that were too long and badly chosen. We took the decision to rename them while
keeping the old classes as aliases of the new ones. This way, old code and documentation
and even tools keep working as expected, until we decide to definitely deprecate
these old classes. For now, they are still defined and available, but we will mark
them as deprecated starting from version 1.3.

A warning will be displayed each time such a class is used, inciting users and
maintainers to update their code to use the new classes, which are strictly
compatible as they expose the same methods and properties. This warning will
include a deadline after which these classes will be definitely removed.

.. autoclass:: whad.device.device.WhadDevice
    :show-inheritance:

.. autoclass:: whad.device.device.WhadVirtualDevice
    :show-inheritance:

.. autoclass:: whad.device.connector.WhadDeviceConnector
    :show-inheritance:
"""

# Load interface base classes
from .device import Device, VirtualDevice, DeviceEvt, Disconnected, MessageReceived, \
    WhadDevice, WhadVirtualDevice
from .connector import Connector, Event, LockedConnector, WhadDeviceConnector
from .bridge import Bridge

# Base device classes
from .uart import Uart
from .tcp import TcpSocket
from .unix import UnixSocket

__all__ = [
    # Base classes
    "Device",
    "VirtualDevice",
    "Connector",
    "WhadDeviceConnector",
    "Bridge",
    "DeviceEvt",
    "Disconnected",
    "MessageReceived",
    "Event",

    # Kept for compatibility
    "WhadDevice",
    "WhadVirtualDevice",
    "WhadDeviceConnector",
    "LockedConnector",

    # Base devices
    "TcpSocket",
    "Uart",
    "UnixSocket",
]



from whad.unifying.connector import Unifying, Sniffer, Keylogger, Mouselogger, Mouse, Keyboard, Dongle, Injector
from whad.unifying.utils.phy import PHYS

__all__ = [
    'Unifying',
    'Sniffer',
    'Keylogger',
    'Mouselogger',
    'Mouse',
    'Keyboard',
    'Injector',
    'Dongle',
    'PHYS'
]



"""
Bluetooth LE Stack Link-layer Manager
"""
import logging
from struct import pack
from random import randint
from threading import Lock
from typing import Tuple

from scapy.layers.bluetooth4LE import *

from whad.common.stack import Layer, alias, source, state, LayerState, instance
from whad.ble.stack.l2cap import L2CAPLayer
from whad.ble.crypto import LinkLayerCryptoManager, generate_random_value, e
from whad.hub.ble import BDAddress

logger = logging.getLogger(__name__)

CONNECTION_UPDATE_REQ = 0x00
CHANNEL_MAP_REQ = 0x01
TERMINATE_IND = 0x02
ENC_REQ = 0x03
ENC_RSP = 0x04
START_ENC_REQ = 0x05
START_ENC_RSP = 0x06
UNKNOWN_RSP = 0x07
FEATURE_REQ = 0x08
FEATURE_RSP = 0x09
PAUSE_ENC_REQ = 0x0A
PAUSE_ENC_RSP = 0x0B
VERSION_IND = 0x0C
REJECT_IND = 0x0D
SLAVE_FEATURE_REQ = 0x0E
CONNECTION_PARAM_REQ = 0x0F
CONNECTION_PARAM_RSP = 0x10
REJECT_IND_EXT = 0x11
PING_REQ = 0x12
PING_RSP = 0x13
LENGTH_REQ = 0x14
LENGTH_RSP = 0x15

class BleConnection:
    """BLE connection implementation
    """

    def __init__(self, l2cap_instance, conn_handle, local_peer_addr, remote_peer_addr):
        self.__l2cap = l2cap_instance
        self.__conn_handle = conn_handle
        self.__remote_peer = remote_peer_addr
        self.__local_peer = local_peer_addr
        self.__lock = Lock()

    @property
    def remote_peer(self):
        """Remote peer address
        """
        return self.__remote_peer

    @property
    def local_peer(self):
        """Local peer address
        """
        return self.__local_peer

    @property
    def conn_handle(self) -> int:
        """Connection handle
        """
        return self.__conn_handle

    @property
    def l2cap(self):
        """Associated L2CAP instance
        """
        return self.__l2cap

    @property
    def gatt(self):
        """Reference to the associated GATT layer
        """
        return self.__l2cap.get_layer("gatt")

    @property
    def smp(self):
        """Security Manager Protocol instance
        """
        return self.__l2cap.get_layer("smp")

    @property
    def phy(self):
        """PHY layer
        """
        return self.__l2cap.get_layer("phy")

    @property
    def ll(self):
        """Link layer
        """
        return self.__l2cap.get_layer("ll")
    
    @property
    def att(self):
        """ATT layer
        """
        return self.__l2cap.get_layer("att")

    @property
    def remote_version(self):
        """Query remote peer BLE version
        """
        return self.ll.state.get_version_remote(self.__conn_handle)

    def lock(self):
        """Lock connection
        """
        self.__lock.acquire()

    def unlock(self):
        """Unlock connection
        """
        self.__lock.release()

    def on_disconnect(self):
        """Connection has been closed.
        """

    def send_version(self):
        """Send LL_VERSION_IND PDU.
        """
        if not self.ll.state.get_connection(self.__conn_handle)['version_sent']:
            # Mark version as sent
            self.ll.state.get_connection(self.__conn_handle)['version_sent'] = True

            # Send LL_VERSION_IND PDU
            self.ll.send_ctrl_pdu(
                self.__conn_handle,
                LL_VERSION_IND(
                    version=self.phy.bt_version.value,
                    company=self.phy.manufacturer_id,
                    subversion=self.phy.bt_sub_version
                )
            )

class LinkLayerState(LayerState):
    """BLE Link layer state
    """

    def __init__(self):
        super().__init__()
        self.connections = {}

    def get_connection(self, conn_handle) -> BleConnection:
        """Retrieve a BleConnection object from a connection handle.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: The BleConnection object associated with a specific connection
                 handle, if any
        :rtype: BleConnection
        :raises: IndexError
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]
        raise IndexError

    def register_connection(self, conn_handle, l2cap_instance, local_peer_addr: BDAddress,
                            remote_peer_addr: BDAddress):
        """Register a connection into this link-layer state.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param l2cap_instance: L2CAP instance
        :type l2cap_instance: Layer
        :param local_peer_addr: Local peer address
        :type local_peer_addr: BDAddress
        :param remote_peer_addr: Remote peer address
        :type remote_peer_addr: BDAddress
        """
        self.connections[conn_handle] = {
            'l2cap': l2cap_instance,
            'local_peer_addr': local_peer_addr.value,
            'local_peer_addr_type': local_peer_addr.type,
            'remote_peer_addr': remote_peer_addr.value,
            'remote_peer_addr_type': remote_peer_addr.type,
            'version_sent': False,  # version exchanged
            'version_remote': None,
            'encryption_key':None,
            'authenticated':False,
            'encrypted':False,
            'skd':None,
            'iv':None,
            'rand':None,
            'ediv':None,
            'nb_pdu_recvd': 0       # number of packets received
        }

    def unregister_connection(self, conn_handle: int):
        """Unregister a specific connection from this state.

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        if conn_handle in self.connections:
            del self.connections[conn_handle]

    def get_connection_l2cap(self, conn_handle: int) -> Layer:
        """Retrieve a connection L2CAP instance

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: L2CAP layer if any, None otherwise
        :rtype: Layer
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]['l2cap']
        return None

    def get_connection_handle(self, l2cap_instance: Layer) -> int:
        """Get connection handle from L2CAP instance

        :param l2cap_instance: L2CAP layer instance
        :type l2cap_instance: Layer
        :return: Associated connection handle if found, else None
        :rtype: None
        """
        for conn_handle, conn in self.connections.items():
            if conn['l2cap'] == l2cap_instance:
                return conn_handle
        return None

    def register_encryption_key(self, conn_handle: int, key: bytes):
        """Register an encryption key for a connection.
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['encryption_key'] = key

    def is_authenticated(self, conn_handle: int) -> bool:
        """Check if connection is authenticated.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: `True` if connection is authenticated, `False` otherwise
        :rtype: bool
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]['authenticated']
        # Not found, return False
        return False

    def is_encrypted(self, conn_handle: int) -> bool:
        """Check if connection is encrypted

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: `True` if connection is encrypted, `False` otherwise
        :rtype: bool
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]['encrypted']
        return False

    def mark_as_authenticated(self, conn_handle: int):
        """Mark connection as authenticated

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['authenticated'] = True

    def mark_as_encrypted(self, conn_handle: int):
        """Mark connection as encrypted

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['encrypted'] = True

    def get_encryption_key(self, conn_handle: int) -> bytes:
        """Retrieve encryption key for a given connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: Encryption key if found, `None` otherwise
        :rtype: bytes
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]['encryption_key']
        return None

    def register_skd_and_iv(self, conn_handle: int, skd: int, iv: int):
        """Register SKD and IV for a given connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param skd: SKD for this connection
        :type skd: int
        :param iv: IV for this connection
        :type iv: int
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['skd'] = skd
            self.connections[conn_handle]['iv'] = iv

    def get_skd_and_iv(self, conn_handle: int) -> Tuple[int, int]:
        """Retrieve SKD and IV for a given connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: SKD and IV if found, tuple of None values otherwise
        :rtype: tuple
        """
        if conn_handle in self.connections:
            skd = self.connections[conn_handle]['skd']
            iv = self.connections[conn_handle]['iv']
            return (skd, iv)
        return (None, None)

    def register_rand_and_ediv(self, conn_handle: int, rand: int, ediv: int):
        """Register RAND and EDIV for a given connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param rand: RAND value
        :type rand: int
        :param ediv: EDIV value
        :type ediv: int
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['rand'] = rand
            self.connections[conn_handle]['ediv'] = ediv

    def get_rand_and_ediv(self, conn_handle: int) -> Tuple[int, int]:
        """Retrieve RAND and EDIV for a given connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: RAND and EDIV values for the given connection
        :rtype: tuple
        """
        if conn_handle in self.connections:
            rand = self.connections[conn_handle]['rand']
            ediv = self.connections[conn_handle]['ediv']
            return (rand, ediv)
        return (None, None)

    def mark_version_sent(self, conn_handle: int):
        """Mark version sent for this connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['version_sent'] = True

    def is_version_sent(self, conn_handle: int) -> bool:
        """Check if version PDU has already been sent

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]['version_sent']
        return False

    def set_version_remote(self, conn_handle: int, version: LL_VERSION_IND):
        """Set version for remote device

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param version: Version information
        :type version: LL_VERSION_IND
        """
        if conn_handle in self.connections:
            self.connections[conn_handle]['version_remote'] = version

    def get_version_remote(self, conn_handle: int) -> LL_VERSION_IND:
        """Retrieve version information for a given connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: Remote peer version PDU if found, `None` otherwise 
        """
        if conn_handle in self.connections:
            return self.connections[conn_handle]['version_remote']
        return None

@alias('ll')
@state(LinkLayerState)
class LinkLayer(Layer):
    """Bluetooth Low Energy link-layer implementation.
    """

    def __init__(self, parent=None, layer_name=None, options=None):
        super().__init__(parent=parent, layer_name=layer_name, options=options)
        self.__llcm = None

    def configure(self, options=None):
        # Control PDU dispatch
        self.__handlers = {
            CONNECTION_UPDATE_REQ: self.on_connection_update_req,
            CHANNEL_MAP_REQ: self.on_channel_map_req,
            TERMINATE_IND: self.on_terminate_ind,
            ENC_REQ: self.on_enc_req,
            ENC_RSP: self.on_enc_rsp,
            START_ENC_REQ: self.on_start_enc_req,
            START_ENC_RSP: self.on_start_enc_rsp,
            UNKNOWN_RSP: self.on_unknown_rsp,
            FEATURE_REQ: self.on_feature_req,
            FEATURE_RSP: self.on_feature_rsp,
            PAUSE_ENC_REQ: self.on_pause_enc_req,
            PAUSE_ENC_RSP: self.on_pause_enc_rsp,
            VERSION_IND: self.on_version_ind,
            REJECT_IND: self.on_reject_ind,
            SLAVE_FEATURE_REQ: self.on_slave_feature_req,
            CONNECTION_PARAM_REQ: self.on_connection_param_req,
            CONNECTION_PARAM_RSP: self.on_connection_param_rsp,
            REJECT_IND_EXT: self.on_reject_ind_ext,
            PING_REQ: self.on_ping_req,
            PING_RSP: self.on_ping_rsp,
            LENGTH_REQ: self.on_length_req,
            LENGTH_RSP: self.on_length_rsp
        }

    def on_connect(self, conn_handle, local_peer_addr, remote_peer_addr):
        """Handles BLE connection
        """
        if conn_handle not in self.state.connections:
            logger.info("[llm] registers new connection %d with %s", conn_handle,
                        remote_peer_addr)

            # Instantiate a L2CAP layer (contextual) to handle the connection
            conn_l2cap = self.instantiate(L2CAPLayer)
            conn_l2cap.set_conn_handle(conn_handle)

            # Update state with new connection
            self.state.register_connection(conn_handle, conn_l2cap.name, local_peer_addr,
                                           remote_peer_addr)

            # Return connection object
            return BleConnection(
                conn_l2cap,
                conn_handle,
                local_peer_addr,
                remote_peer_addr
            )
        # Connection already exists
        logger.error('[!] Connection already exists')

        # Return connection object
        return BleConnection(
            self,
            conn_handle,
            local_peer_addr,
            remote_peer_addr
        )

    def on_disconnect(self, conn_handle: int):
        """Handle disconnection event

        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        # Free the previously instantiated L2CAP layer
        conn_layer = self.state.get_connection_l2cap(conn_handle)
        if conn_layer is not None:
            # Mark GATT layer as disconnected
            self.get_layer(conn_layer).get_layer("gatt").state.terminated = True
            self.destroy(self.get_layer(conn_layer))

        # Remove connection from our registered connections
        self.state.unregister_connection(conn_handle)

    @source('phy', 'data')
    def on_data_pdu_recv(self, pdu: Packet, conn_handle: int = None):
        """Handle data PDU sent by our PHY layer

        :param pdu: Incoming PDU to process
        :type pdu: Packet
        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        # Count PDU
        if conn_handle in self.state.connections:
            conn_metadata = self.state.get_connection(conn_handle)
            conn_metadata['nb_pdu_recvd'] += 1

        # We received a data PDU
        self.on_data_pdu(pdu, conn_handle)

    @source('phy', 'control')
    def on_ctrl_pdu_recv(self, pdu: Packet, tag: str = None, conn_handle: int = None):
        """Handle control PDU received by our PHY layer

        :param pdu: Received control PDU
        :type pdu: Packet
        :param tag: Optional tag
        :type tag: str, optional
        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        # Count PDU
        if conn_handle in self.state.connections:
            conn_metadata = self.state.get_connection(conn_handle)
            conn_metadata['nb_pdu_recvd'] += 1

        # We received a control PDU
        self.on_ctrl_pdu(pdu, conn_handle)

    def on_ctrl_pdu(self, pdu: Packet, conn_handle: int):
        """Process a specific control PDU by calling the associated
        control PDU type handler.

        :param pdu: Control PDU
        :type pdu: Packet
        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        if conn_handle in self.state.connections:
            ctrl = pdu.getlayer(BTLE_CTRL)
            if ctrl.opcode in self.__handlers:
                self.__handlers[int(ctrl.opcode)](conn_handle, ctrl.getlayer(1))
        else:
            logger.error('[!] Unknown connection handle: %d', conn_handle)


    def on_data_pdu(self, pdu: Packet, conn_handle: int):
        """Forward data PDU to the upper layer (L2CAP) instance of the
        connection identified by its connection handle.

        :param pdu: Incoming data PDU
        :type pdu: Packet
        :param conn_handle: Connection handle
        :type conn_handle: int
        """
        # We look for the corresponding L2CAP layer instance
        l2cap_layer = self.state.get_connection_l2cap(conn_handle)
        if l2cap_layer is not None:
            self.send(l2cap_layer, bytes(pdu.payload), fragment=pdu.LLID == 0x1)

    @instance('l2cap')
    def on_l2cap_send_data(self, l2cap_inst: Layer, data: Packet, fragment=False, encrypt=None):
        '''Handle L2CAP packets coming from our L2CAP layer and encapsulate them into a
        BTLE_DATA packet (data PDU).

        :param l2cap_inst: Connection L2CAP instance
        :type l2cap_inst: Layer
        :param data: L2CAP packet to forward to PHY
        :type data: Packet
        :param fragment: If set to `True`, will set the fragment flag in the data PDU.
        :type fragment: bool
        :param encrypt: If set to `True`, enable encryption for this data PDU
        :type encrypt: bool
        '''
        # Retrieve connection handle corresponding to the instance
        conn_handle = self.state.get_connection_handle(l2cap_inst)
        if conn_handle is not None:
            logger.debug("sending l2cap data PDU for conn_handle %d (%d bytes)", conn_handle,len(data))
            llid = 0x01 if fragment else 0x02
            self.send(
                'phy',
                BTLE_DATA(
                    LLID=llid,
                    len=len(data)
                )/data,
                tag='data',
                conn_handle=conn_handle,
                encrypt=encrypt
            )
        else:
            logger.error("no connection handle found for L2CAP instance %s", instance)

    @instance('l2cap', tag='ATT_MTU')
    def on_l2cap_mtu_changed(self, l2cap_inst: Layer, mtu: int):
        """Handle MTU value change notification from L2CAP layer.

        :param l2cap_inst: Connection L2CAP instance
        :type l2cap_inst: Layer
        :param mtu: Updated MTU value
        :param mtu: int
        """
        # Retrieve connection handle corresponding to the instance
        conn_handle = self.state.get_connection_handle(l2cap_inst)
        if conn_handle is not None:
            logger.debug("sending updated ATT MTU (%d) for conn_handle %d", mtu, conn_handle)
            self.send(
                'phy',
                mtu,
                tag='ATT_MTU',
                conn_handle=conn_handle
            )

    def send_ctrl_pdu(self, conn_handle: int, pdu: Packet, encrypt: bool = None):
        """Send a control PDU to the underlying PHY layer.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param pdu: control PDU to send
        :type pdu: Packet
        :param encrypt: If set to `True`, control PDU will be sent encrypted
        :type encrypt: bool
        """
        self.send('phy', BTLE_DATA()/BTLE_CTRL()/pdu, tag='control',
                  conn_handle=conn_handle, encrypt=encrypt)


    ### Link-layer control PDU callbacks

    def on_unsupported_opcode(self, conn_handle: int, opcode: int):
        """Handle unsupported control PDU opcodes.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param opcode: control PDU opcode
        :type opcode: int
        """
        self.send_ctrl_pdu(
            conn_handle,
            LL_UNKNOWN_RSP(code=opcode)
        )

    def on_connection_update_req(self, conn_handle: int, conn_update: LL_CONNECTION_UPDATE_IND):
        """Connection update is not supported yet

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param conn_update: Connection update PDU
        :type conn_update: LL_CONNECTION_UPDATE_IND
        """
        self.on_unsupported_opcode(conn_handle, CONNECTION_UPDATE_REQ)

    def on_channel_map_req(self, conn_handle: int, channel_map: LL_CHANNEL_MAP_IND):
        """Channel map update is not supported yet

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param channel_map: channel map PDU
        :type channel_map: LL_CHANNEL_MAP_IND
        """
        self.on_unsupported_opcode(conn_handle, CHANNEL_MAP_REQ)

    def on_terminate_ind(self, conn_handle: int, terminate: LL_TERMINATE_IND):
        """Terminate this connection

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param terminate: Terminate PDU
        :type terminate: LL_TERMINATE_IND
        """
        # Connection has been terminated
        conn = self.state.get_connection(conn_handle)
        if conn is not None:
            self.on_disconnect(conn_handle)

    def start_encryption(self, conn_handle: int, rand: int, ediv: int):
        """
        Initiate encryption procedure.
        """
        # Retrieve connection handle corresponding to the instance

        encryption_key = None
        if conn_handle is not None:
            encryption_key = self.state.get_encryption_key(conn_handle)

        # Allowed if we have already negociated an STK
        if encryption_key is not None and conn_handle is not None:

            # Generate our SKD and IV
            skd = randint(0, 0x10000000000000000)
            iv = randint(0, 0x100000000)
            self.state.register_skd_and_iv(conn_handle, skd, iv)

            logger.info('[llm] Initiate connection LinkLayerCryptoManager')

            # Save master rand/iv
            self.state.register_rand_and_ediv(conn_handle, rand, ediv)

            # Send back our parameters
            self.send_ctrl_pdu(
                conn_handle,
                LL_ENC_REQ(
                    rand = rand,
                    ediv = ediv,
                    skdm = skd,
                    ivm = iv
                )
            )

        else:
            self.send_ctrl_pdu(
                conn_handle,
                LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )

    def on_enc_rsp(self, conn_handle, enc_rsp: LL_ENC_RSP):
        """Encryption response handler

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param enc_rsp: Encryption response PDU
        :type enc_rsp: LL_ENC_RSP
        """
        # Retrieve connection handle corresponding to the instance

        encryption_key = None
        if conn_handle is not None:
            encryption_key = self.state.get_encryption_key(conn_handle)


        # Allowed if we have already negociated an STK
        if encryption_key is not None and conn_handle is not None:

            skdm, ivm = self.state.get_skd_and_iv(conn_handle)
            rand, ediv = self.state.get_rand_and_ediv(conn_handle)

            logger.info(
                "[llm] Received LL_ENC_RSP: skds=%s ivs=%s",
                pack('<Q', enc_rsp.skds).hex(),
                pack('<I', enc_rsp.ivs).hex(),
            )

            logger.info('[llm] Initiate connection LinkLayerCryptoManager')

            # Initiate LLCM
            self.__llcm = LinkLayerCryptoManager(
                encryption_key,
                skdm,
                ivm,
                enc_rsp.skds,
                enc_rsp.ivs
            )

            logger.debug("[llm] crypto material saved, waiting for an LL_START_ENC_REQ")
            """
            # Notify encryption enabled
            if not self.get_layer('phy').set_encryption(
                conn_handle=conn_handle,
                enabled=True,
                ll_key=session_key,
                ll_iv=iv,
                key=encryption_key,
                rand=rand,
                ediv=ediv
            ):
                logger.info('[llm] Cannot enable encryption')
            else:
                logger.info('[llm] Encryption enabled in hardware')
            """
        else:
            self.send_ctrl_pdu(
                conn_handle,
                LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )


    def on_enc_req(self, conn_handle: int, enc_req: LL_ENC_REQ):
        """Encryption request handler

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param enc_req: Encryption request PDU
        :type enc_req: LL_ENC_REQ
        """
        # Retrieve connection handle corresponding to the instance

        encryption_key = None
        if conn_handle is not None:
            encryption_key = self.state.get_encryption_key(conn_handle)

        # Allowed if we have already negociated an STK
        if encryption_key is not None and conn_handle is not None:

            # Generate our SKD and IV
            skd = randint(0, 0x10000000000000000)
            iv = randint(0, 0x100000000)
            self.state.register_skd_and_iv(conn_handle, skd, iv)

            logger.info(
                "[llm] Received LL_ENC_REQ: rand=%s ediv=%s skd=%s iv=%s",
                pack('<Q', enc_req.rand).hex(),
                pack('<H', enc_req.ediv).hex(),
                pack('<Q', enc_req.skdm).hex(),
                pack('<I', enc_req.ivm).hex(),
            )

            logger.info("[llm] Initiate connection LinkLayerCryptoManager")

            # Save master rand/iv
            self.state.register_rand_and_ediv(conn_handle, enc_req.rand, enc_req.ediv)

            # Initiate LLCM
            self.__llcm = LinkLayerCryptoManager(
                encryption_key,
                enc_req.skdm,
                enc_req.ivm,
                skd,
                iv
            )

            # Compute session key
            master_skd = pack(">Q", enc_req.skdm)
            master_iv = pack("<L", enc_req.ivm)
            slave_skd = pack(">Q", skd)
            slave_iv = pack("<L", iv)

            # Generate session key diversifier
            skd = slave_skd + master_skd

            # Generate initialization vector
            iv = master_iv + slave_iv

            # Generate session key
            session_key = e(encryption_key, skd)

            logger.info("[llm] master  skd: %s", master_skd.hex())
            logger.info("[llm] master   iv: %s", master_iv.hex())
            logger.info("[llm] slave   skd: %s", slave_skd.hex())
            logger.info("[llm] slave    iv: %s", slave_iv.hex())
            logger.info("[llm] Session  TK: %s", encryption_key.hex())
            logger.info("[llm] Session  iv: %s", iv.hex())
            logger.info("[llm] Exp. Ses iv: %s", self.__llcm.iv.hex())
            logger.info("[llm] Session key: %s", session_key.hex())
            skdm, ivm = self.state.get_skd_and_iv(conn_handle)
            logger.info(
                "[llm] Send LL_ENC_RSP: skd=%s iv=%s",
                pack('<Q', skdm).hex(),
                pack('<I', ivm).hex(),
            )

            # Send back our parameters
            self.send_ctrl_pdu(
                conn_handle,
                LL_ENC_RSP(
                    skds = skdm,
                    ivs = ivm
                )
            )

            # Notify encryption enabled
            if not self.get_layer('phy').set_encryption(
                conn_handle=conn_handle,
                enabled=True,
                ll_key=session_key,
                ll_iv=iv,
                key=encryption_key,
                rand=enc_req.rand,
                ediv=enc_req.ediv
            ):
                logger.info('[llm] Cannot enable encryption')
            else:
                logger.info('[llm] Encryption enabled in hardware')

            # Start encryption (STK as LTK)
            self.send_ctrl_pdu(
                conn_handle,
                LL_START_ENC_REQ(),
                encrypt=False
            )

        else:
            self.send_ctrl_pdu(
                conn_handle,
                LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )

    def on_start_enc_req(self, conn_handle: int, start_enc_req: LL_START_ENC_REQ):
        """Encryption start request handler.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param start_enc_req: Start encryption request PDU
        :type start_enc_req: LL_START_ENC_REQ
        """
        logger.debug("[llm] Received LL_START_ENC_REQ")
        # Compute session key
        #master_skd = pack(">Q", self.__llcm.master_skd)
        #master_iv = pack("<L", self.__llcm.mast)
        #slave_skd = pack(">Q", enc_rsp.skds)
        #slave_iv = pack("<L", enc_rsp.ivs)

        # Generate session key diversifier
        skd = self.__llcm.slave_skd + self.__llcm.master_skd

        # Generate initialization vector
        iv = self.__llcm.master_iv + self.__llcm.slave_iv

        # Generate session key
        session_key = e(self.__llcm.ltk, skd)

        logger.info("[llm] master  skd: %s", self.__llcm.master_skd.hex())
        logger.info("[llm] master   iv: %s", self.__llcm.master_iv.hex())
        logger.info("[llm] slave   skd: %s", self.__llcm.slave_skd.hex())
        logger.info("[llm] slave    iv: %s", self.__llcm.slave_iv.hex())
        logger.info("[llm] Session  TK: %s", self.__llcm.ltk.hex())
        logger.info("[llm] Session  iv: %s", iv.hex())
        logger.info("[llm] Session key: %s", session_key.hex())

        # Start encryption (STK as LTK)
        self.send_ctrl_pdu(
            conn_handle,
            LL_START_ENC_RSP(),
            encrypt=False,
        )

        rand, ediv = self.state.get_rand_and_ediv(conn_handle)

        # Notify encryption enabled
        if not self.get_layer('phy').set_encryption(
            conn_handle=conn_handle,
            enabled=True,
            ll_key=session_key,
            ll_iv=iv,
            key=self.__llcm.ltk,
            rand=rand,
            ediv=ediv
        ):
            logger.info('[llm] Cannot enable encryption')
        else:
            logger.info('[llm] Encryption enabled in hardware')

    def on_start_enc_rsp(self, conn_handle: int, start_enc_rsp: LL_START_ENC_RSP):
        """Encryption start response handler

        Normally, we get this packet when a link has successfully
        been encrypted (with STK or LTK). So we need to notify the
        SMP that encryption has been acknowledged by the remote peer.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param start_enc_rsp: Start encryption response PDU
        :type start_enc_rsp: LL_START_ENC_RSP
        """
        # Check if we are the encryption initiator,
        # if yes then we need to answer to this encrypted LL_START_ENC_RSP
        # with another encrypted LL_START_ENC_RSP
        l2cap_instance = self.state.get_connection_l2cap(conn_handle)
        if not self.get_layer(l2cap_instance).get_layer('smp').is_initiator():
            self.send_ctrl_pdu(
                conn_handle,
                LL_START_ENC_RSP()
            )

        #print("Marked as encrypted")
        self.state.mark_as_encrypted(conn_handle)

        # Notify SMP channel is now encrypted
        self.get_layer(l2cap_instance).get_layer('smp').on_channel_encrypted()

    def on_unknown_rsp(self, conn_handle: int, unk_rsp: LL_UNKNOWN_RSP):
        """handle unknown response PDU (not expected).

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param unk_rsp: Unknown response PDU
        :type unk_rsp: LL_UNKNOWN_RSP
        """

    def on_feature_req(self, conn_handle: int, feature_req: LL_FEATURE_REQ):
        """Handle feature request

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param feature_req: Feature request PDU
        :type feature_req: LL_FEATURE_REQ
        """
        #self.on_unsupported_opcode(FEATURE_REQ)
        # Reply with our basic feature set
        self.send_ctrl_pdu(
            conn_handle,
            LL_FEATURE_RSP(feature_set=[
                'le_encryption',
                'le_ping'
            ])
        )

    def on_feature_rsp(self, conn_handle: int, feature_rsp: LL_FEATURE_RSP):
        """Features not supported yet

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param feature_rsp: Feature response PDU
        :type feature_rsp: LL_FEATURE_RSP
        """
        self.on_unsupported_opcode(conn_handle, FEATURE_RSP)

    def on_pause_enc_req(self, conn_handle: int, pause_enc_req: LL_PAUSE_ENC_RSP):
        """Encryption not supported yet.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param pause_enc_req: Pause encryption request PDU
        :type pause_enc_req: LL_PAUSE_ENC_RSP
        """
        self.on_unsupported_opcode(conn_handle, PAUSE_ENC_REQ)

    def on_pause_enc_rsp(self, conn_handle: int, pause_enc_rsp: LL_PAUSE_ENC_RSP):
        """Encryption not supported yet.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param pause_enc_rsp: Pause encryption response PDU
        :type pause_enc_rsp: LL_PAUSE_ENC_RSP
        """
        self.on_unsupported_opcode(conn_handle, PAUSE_ENC_RSP)

    def on_version_ind(self, conn_handle: int, version: LL_VERSION_IND):
        """Send back our version info

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param version: Remote peer version PDU
        :type version: LL_VERSION_IND
        """
        logger.debug('received a VERSION_IND PDU')
        if not self.state.is_version_sent(conn_handle):
            logger.debug('sending back our VERSION_IND PDU')

            # send control PDU
            self.send_ctrl_pdu(
                conn_handle,
                LL_VERSION_IND(
                    version=self.get_layer('phy').bt_version.value,
                    company=self.get_layer('phy').manufacturer_id,
                    subversion=self.get_layer('phy').bt_sub_version
                )
            )
        else:
            logger.debug('VERSION_IND PDU already sent, skip.')
        self.state.set_version_remote(conn_handle, version)
        self.state.mark_version_sent(conn_handle)

    def on_reject_ind(self, conn_handle: int, reject: int):
        """Handle reject PDU

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param reject: rejection reason
        :type reject: int
        """

    def on_slave_feature_req(self, conn_handle: int, feature_req: LL_SLAVE_FEATURE_REQ):
        """Handle slave feature request, return unsupported PDU for now.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param feature_req: Slave feature request packet
        :type feature_req: Packet
        """
        self.on_unsupported_opcode(conn_handle, SLAVE_FEATURE_REQ)

    def on_connection_param_req(self, conn_handle: int, conn_param_req: LL_CONNECTION_PARAM_REQ):
        """Handle connection parameter request

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param conn_param_req: Connection parameter request packet
        :type conn_param_req: LL_CONNECTION_PARAM_REQ
        """
        self.on_unsupported_opcode(conn_handle, CONNECTION_PARAM_REQ)

    def on_connection_param_rsp(self, conn_handle: int, conn_param_rsp: LL_CONNECTION_PARAM_RSP):
        """Handle connection parameter response

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param conn_param_rsp: Connection parameter response
        :type conn_param_rsp: LL_CONNECTION_PARAM_RSP
        """

    def on_reject_ind_ext(self, conn_handle: int, reject_ext: LL_REJECT_EXT_IND):
        """Handle extended rejection PDU

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param reject_ext: Extended rejection PDU
        :type reject_ext: LL_REJECT_EXT_IND
        """

    def on_ping_req(self, conn_handle: int, ping_req: LL_PING_REQ):
        """Handle ping request (not supported for now)

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param ping_req: Ping request PDU
        :type ping_req: LL_PING_REQ
        """
        self.on_unsupported_opcode(conn_handle, PING_REQ)

    def on_ping_rsp(self, conn_handle: int, ping_rsp: LL_PING_RSP):
        """Handle ping response (unsupported for now).

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param ping_rsp: Ping response PDU
        :type ping_rsp: LL_PING_RSP
        """

    def on_length_req(self, conn_handle: int, length_req: LL_LENGTH_REQ):
        """Received a length request PDU (unsupported for now)

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param length_req: Length request PDU
        :type length_req: LL_LENGTH_REQ
        """
        self.on_unsupported_opcode(conn_handle, LENGTH_REQ)

    def on_length_rsp(self, conn_handle: int, length_rsp: LL_LENGTH_RSP):
        """Handle length response

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param length_rsp: Length response
        :type length_rsp: LL_LENGTH_RSP
        """

LinkLayer.add(L2CAPLayer)
